-- "Hexanimals"
-- By Tim Sandford, 2020
-- Comments to timsandford42@hotmail.com

require("draw42")
require("file42")
require("hex42")
require("math42")
require("music42")
require("table42")

function love.run()   -- Main function
  gVERSION, gPATCH, gREVISION, gCODENAME = love.getVersion()
  if gVERSION >=  11 then  -- Colour fix for 11.0 onwards
    local oldSetColor = love.graphics.setColor
    love.graphics.setColor = function (r, g, b, a)     
      if type(r)=="table" then
        r, g, b, a = r[1] / 255, r[2] / 255, r[3] / 255, (r[4] or 255) / 255
      else
        r, g, b, a = r / 255, g / 255, b / 255, (a or 255) / 255
      end
      oldSetColor(r,g,b,a)
    end
  end
  math.randomseed(os.time())  -- True randomness
  love.load(love.arg.parseGameArguments(arg), arg) -- load variables etc.
  while true do  -- Main loop
    if love.event then -- Process events
      love.event.pump()
      for name, a,b,c,d,e,f in love.event.poll() do
        if name == "quit" then
          if not love.quit or not love.quit() then
            return a
          end
        end
        love.handlers[name](a,b,c,d,e,f)
      end
    end
    love.update() -- Update
    love.graphics.origin() -- Not sure, probably reset origin?
    love.graphics.clear(love.graphics.getBackgroundColor()) -- Clears the screen
    love.draw() -- Collect draw commands
    love.graphics.present() -- Present the graphics on the screen!
    love.timer.sleep(0.001) -- This is to lower the load on the CPU, but I could edit it :)	
  end
end

function love.load(arg)  -- Load everything in
  local i, dir, file
  io.stdout:setvbuf("no")                                         -- Print messages immediately, allows tracing bugs
  if arg[#arg] == "-debug" then require("mobdebug").start() end   -- Zerobrane debugging
  gSCREENWIDTH = love.graphics.getWidth()                         -- Get screensize
  gSCREENHEIGHT = love.graphics.getHeight()                       -- Get screensize
  gRATIOX = gSCREENWIDTH / 800                                    -- Scale graphics to screen
  gRATIOY = gSCREENHEIGHT / 600                                   -- Scale graphics to screen
  gHEX_WIDTH, gHEX_HEIGHT = 120, 140                              -- Size of a hex tile
  gFADE, gFADESPEED = 255, 5                                      -- Fade in effect
  gTIMER = love.timer.getTime()                                   -- Use timer in love.update for speed of game
  gFLASH, gFLASH_z, gFLASH_min, gFLASH_max = 255, 10, 80, 255		  -- Flashing images/text transparency
  gGAMEMODE = "title"                                             -- Gamemode
  gENDING = 0                                                     -- Countdown to the end of the game
  gMOUSEGRAB = 0                                                  -- Count for time grabbing mouse
  gSCROLLCOUNTER = 0                                              -- Count for screen scrolling
  gPOPUP_COUNT, gPOPUP_MSG = 0, ""                                -- Pop up messages, one at a time
  gLEVEL, gDIFFICULTY = 0, 0                                      -- Used in campaign
  AIsimulation = false                                            -- When true, actions are reset

  love.mouse.setCursor(love.mouse.newCursor("gui/cursor.png", 4, 4))
  love.mouse.setVisible(true)

  fTINY = love.graphics.newFont("fonts/regular.ttf", 8)           -- tiny font
  fSMALL = love.graphics.newFont("fonts/heading.ttf", 12)         -- small font
  fDEFAULT = love.graphics.newFont("fonts/heading.ttf", 16)       -- default font
  fSUBHEADING = love.graphics.newFont("fonts/heading.ttf", 24)    -- sub heading font
  fHEADING = love.graphics.newFont("fonts/heading.ttf", 32)       -- heading font
  fTITLE = love.graphics.newFont("fonts/heading.ttf", 50)         -- title font

  iICON, iMAP, iIMAGE = {}, {}, {}
  sSOUND = {}
  for i = 1, 99, 1 do
    if love.filesystem.getInfo("gui/icon"..i..".png", "file") then iICON[i] = love.graphics.newImage("gui/icon"..i..".png") end
    if love.filesystem.getInfo("map/map"..i..".jpg", "file") then iMAP[i] = love.graphics.newImage("map/map"..i..".jpg") end
    if love.filesystem.getInfo("images/pic"..i..".jpg", "file") then iIMAGE[i] = love.graphics.newImage("images/pic"..i..".jpg") end
    if love.filesystem.getInfo("sounds/sound"..i..".ogg", "file") then sSOUND[i] = love.audio.newSource("sounds/sound"..i..".ogg", "static") end
  end
  iUNKNOWN = love.graphics.newImage("gui/unknown.png")
  iBACKGROUND = love.graphics.newImage("backgrounds/background0.jpg")
  iFRONT = love.graphics.newImage("gui/CardFront.png")
  iSQUARE = love.graphics.newImage("gui/CardSquare.png")
  iBACK = love.graphics.newImage("gui/CardBack.png")

  xHAND = {590, 590}
  yHAND = {520, 60}
  xDECK = {60, 60}
  yDECK = {530, 80}
  xGRAVE = {740, 740}
  yGRAVE = {360, 240}
  xBUTTONS = {355, 355}
  yBUTTONS = {506, 67}
  xGRID, yGRID = 200, 300

  if love.filesystem.getInfo("settings.txt", "file") then tSETTINGS = readsingletable("settings.txt")
  else tSETTINGS = readsingletable("data/defaultsettings.txt") end

  if love.filesystem.getInfo("player.txt", "file") then tPLAYER = readsingletable("player.txt")
  else tPLAYER = readsingletable("data/newplayer.txt") end

  tCOLOUR = readmultitable("data/colours.txt")
  tMAP = readmultitable("data/map.txt")
  tBLUEPRINT = readmultitable("data/cards.txt")
  tABILITIES = readmultitable("data/abilities.txt")
  tEFFECTS = readmultitable("data/effects.txt")
  tDECK = {}

  if love.filesystem.getInfo("collection.txt", "file") then tCOLLECTION = loaddeck("collection.txt")
  else tCOLLECTION = loaddeck("data/defaultdeck.txt") end

  if not love.filesystem.getInfo("decks", "directory") then love.filesystem.createDirectory("decks") end
  
  if love.filesystem.getInfo("deck.txt", "file") then tDECK[1] = loaddeck("deck.txt")
  else tDECK[1] = loaddeck("data/defaultdeck.txt") end

  savedeck(tDECK[1], "deck.txt")
  savedeck(tCOLLECTION, "collection.txt")
  musicload(tSETTINGS.MusicVol)
  playtrack(1)
end

function savedeck(deck, filename)  -- Save the deck
  local files = getordereddirectory("decks")
  local list = {}
  local i
  if #files<=25 then
    if next(deck)~=nil then
      for i = 1, #deck, 1 do
        list[i] = deck[i].Name
      end
    end
    writesinglelist(filename, list)
  else addpopup("Directory Full") end
end

function loaddeck(filename)  -- Load a deck, returns as a list
  local list, i, j
  deck = {}
  if love.filesystem.getInfo(filename) then
    list = readsinglelist(filename)
    if next(list)~=nil and next(tBLUEPRINT)~=nil then
      for i = 1, #list, 1 do
        for j = 1, #tBLUEPRINT, 1 do
          if tBLUEPRINT[j].Name~=nil and tBLUEPRINT[j].Name==list[i] then table.insert(deck, deepcopy(tBLUEPRINT[j])) end
        end
      end
    end
  end
  deck = getlistofcards(deck)
  return deck
end

function newgame()  -- Shuffle decks, setup and deal initial cards
  local i	, j
  writelog(os.date("%Y-%m-%d %H:%M\r\n"))
  writelog("NEW GAME: "..tCONTROL[1].." vs "..tCONTROL[2].."\r\n")
  gTURN = math.random(2)
  gDAYNIGHT = math.random(10)
  gCEASEFIRE = 0
  gBONUSCARD, gNEWLEVEL = 0, 0
  if gLEVEL>=1 and gLEVEL<=11 then iBACKGROUND = love.graphics.newImage("backgrounds/background"..gLEVEL..".jpg")
  else iBACKGROUND = love.graphics.newImage("backgrounds/background"..math.random(11)..".jpg") end
  blankgrid()
  tHAND, tCOMBAT, tGRAVE = {}, {}, {}
  tBUDGET = {0, 0}
  tPOINTS = {0, 0}
  for j = 1, 2, 1 do
    tHAND[j], tCOMBAT[j], tGRAVE[j] = {}, {}, {}
    for i = 1, #tDECK[j], 1 do
      blankcard(tDECK[j][i], "Yes")
      tDECK[j][i].X = xDECK[j]
      tDECK[j][i].Y = yDECK[j]
      if j == 1 then
        tDECK[j][i].R = tCOLOUR[tSETTINGS.P1COLOUR].R
        tDECK[j][i].G = tCOLOUR[tSETTINGS.P1COLOUR].G
        tDECK[j][i].B = tCOLOUR[tSETTINGS.P1COLOUR].B
      else
        tDECK[j][i].R = tCOLOUR[tSETTINGS.P2COLOUR].R
        tDECK[j][i].G = tCOLOUR[tSETTINGS.P2COLOUR].G
        tDECK[j][i].B = tCOLOUR[tSETTINGS.P2COLOUR].B
      end
      tDECK[j][i].FaceUp = false
      tDECK[j][i].Clickable = false
      tDECK[j][i].Animate = false
      tDECK[j][i].AniCount = 0
      tDECK[j][i].OriginalX = xDECK[j]
      tDECK[j][i].OriginalY = yDECK[j]
      tDECK[j][i].TargetX = xDECK[j]
      tDECK[j][i].TargetY = yDECK[j]
      tDECK[j][i].FloatyText = {}
    end
  end
  playsound(2)
  for i = 1, tSETTINGS.Shuffle, 1 do
    shuffledeck(1)
    shuffledeck(2)
  end
  for j = 1, 2, 1 do
    for i = 1, 3, 1 do
      dealcard(j)
    end
    tBUDGET[j] = 0
    tPOINTS[j] = 0
  end
  dealcard(3-gTURN)
  arrangehand(3-gTURN)
  newturn(gTURN)
  playtrack(3)
end

function blankgrid()
  local i, j, row_width, mid_point
  if tCONTROL[1]=="AI" or tCONTROL[2]=="AI" then gGRID_WIDTH, gGRID_HEIGHT = 3, 5
  elseif tSETTINGS.Gridsize=="Small" then gGRID_WIDTH, gGRID_HEIGHT = 3, 5
  elseif tSETTINGS.Gridsize=="Medium" then gGRID_WIDTH, gGRID_HEIGHT = 4, 7
  else gGRID_WIDTH, gGRID_HEIGHT = 5, 9 end
  gGRID_MID = math.ceil((gGRID_WIDTH+1)/2)
  tHEX = {}
  for j = 1, 9, 1 do
    tHEX[j] = {}
    for i = 1, 5, 1 do
      tHEX[j][i] = {}
      tHEX[j][i].Player = -1
      tHEX[j][i].Moves = math.huge
    end
  end
  for j = 1, gGRID_HEIGHT, 1 do
    if j<=gGRID_WIDTH then row_width = j
    else row_width = (2*gGRID_WIDTH)-j end
    for i = gGRID_MID-math.ceil((row_width-1)/2), gGRID_MID+math.ceil((row_width-1)/2)-math.abs((j%2)-1), 1 do
      tHEX[j][i].Player = 0
    end
  end
  tHEX[1][gGRID_MID-1].Player = 0
  tHEX[1][gGRID_MID+1].Player = 0
  tHEX[gGRID_HEIGHT][gGRID_MID-1].Player = 0
  tHEX[gGRID_HEIGHT][gGRID_MID+1].Player = 0
end

function shuffledeck(pl)
  local i = math.random(1, #tDECK[pl])
  local j = math.random(1, #tDECK[pl])
  tDECK[pl][i], tDECK[pl][j] = deepcopy(tDECK[pl][j]), deepcopy(tDECK[pl][i])
end

function blankcard(card, heal)  -- Used in newgame but also a couple of actions
  local i
  card.FaceUp = true
  card.Steps = 0
  for i = 1, 4, 1 do
    card["Charge"..i]=0
  end
  card.Target = false
  card.X, card.Y = 0, 0
  card.GX, card.GY = 0, 0
  card.Angle = 0
  card.Animate = false
  card.AniCount = 0
  card.OriginalX, card.OriginalY = 0, 0
  card.TargetX, card.TargetY = 0, 0
  card.Effect = {}
  for i = 1, #tEFFECTS, 1 do
    card.Effect[i] = 0
  end
  if heal == "Yes" then card.Wounds = 0 end
end

function dealcard(pl)
  local size = #tHAND[pl]
  if next(tDECK[pl]) ~= nil then
    sendcard("Hand", tDECK[pl][1], "Deck", pl, 1)
  end
end

function arrangehand(pl)
  local width, height = tSETTINGS.HandWidth-tSETTINGS.CardWidth, tSETTINGS.HandHeight-tSETTINGS.CardHeight
  local i, j, m
  if next(tHAND[pl])~=nil then
    m = (#tHAND[pl]+1)/2
    for i = 1, #tHAND[pl], 1 do
      j = (i-m)/m
      tHAND[pl][i].OriginalX = tHAND[pl][i].X
      tHAND[pl][i].OriginalY = tHAND[pl][i].Y
      tHAND[pl][i].TargetX = xHAND[pl]+(j*width)
      tHAND[pl][i].TargetY = yHAND[pl]+height-(math.cos(j)*height)
      tHAND[pl][i].Angle = j*tSETTINGS.HandSlant
      tHAND[pl][i].Animate = true
      tHAND[pl][i].AniCount = tSETTINGS.AniTime
    end
  end
end

function newturn(pl)  -- deal cards, temporary effects end
  local a, b = pl, 3 - pl     -- friendly & foe
  local i, j
  local spaces, removed
  writelog("\r\nP"..a..": NEW TURN")
  if pl==1 then gDAYNIGHT = gDAYNIGHT + 1 end
  if gDAYNIGHT>10 then gDAYNIGHT = 1 end
  gCEASEFIRE = math.max(gCEASEFIRE-1, 0)
  if #tHAND[a] < tSETTINGS.MaxCardsHand then dealcard(a) end
  if #tHAND[a] < tSETTINGS.MaxCardsHand then dealcard(a) end
  tPOINTS[a] = tBUDGET[a]
  if next(tDECK[a]) ~= nil then
    for i = #tDECK[a], 1, -1 do
      tDECK[a][i].FaceUp = false
      tDECK[a][i].Clickable = false
      for j = 1, #tEFFECTS, 1 do
        if tDECK[a][i].Effect[j]<100 then tDECK[a][i].Effect[j] = math.max(tDECK[a][i].Effect[j]-1, 0) end
      end
    end
  end
  if next(tDECK[b]) ~= nil then
    for i = #tDECK[b], 1, -1 do
      tDECK[b][i].FaceUp = false
      tDECK[b][i].Clickable = false
      for j = 1, #tEFFECTS, 1 do
        if tDECK[b][i].Effect[j]<100 then tDECK[b][i].Effect[j] = math.max(tDECK[b][i].Effect[j]-1, 0) end
      end
    end    
  end
  if next(tCOMBAT[a]) ~= nil then
    for i = #tCOMBAT[a], 1, -1 do
      removed = false
      tCOMBAT[a][i].FaceUp = true
      tCOMBAT[a][i].Clickable = true
      tCOMBAT[a][i].Steps = 0
      for j = 1, 4, 1 do
        if tCOMBAT[a][i]["Spec"..j]~=nil then tCOMBAT[a][i]["Charge"..j] = math.min(tCOMBAT[a][i]["Charge"..j]+1, getrechargetime(tCOMBAT[a][i]["Spec"..j])) end
      end
      for j = 1, #tEFFECTS, 1 do
        if tCOMBAT[a][i].Effect[j]<100 then tCOMBAT[a][i].Effect[j] = math.max(tCOMBAT[a][i].Effect[j]-1, 0) end
      end
      if gethasability(tCOMBAT[a][i], "Nocturnal") then
        if gDAYNIGHT>=6 then tCOMBAT[a][i].Effect[10] = 1
        else tCOMBAT[a][i].Effect[9] = 1 end
      end
      if gethasability(tCOMBAT[a][i], "Lazy") and math.random(100)<=15 then
        tCOMBAT[a][i].Effect[1] = math.max(tCOMBAT[a][i].Effect[1], 1)
        addfloatytext(tCOMBAT[a][i], "-Lazy-")
      end
      if gethasability(tCOMBAT[a][i], "Slow") and math.random(100)<=25 then
        tCOMBAT[a][i].Effect[1] = math.max(tCOMBAT[a][i].Effect[1], 1)
        addfloatytext(tCOMBAT[a][i], "-Slow-")
      end
      if gethasability(tCOMBAT[a][i], "Breeds") and math.random(100)<=15 then
        spaces = getfreespaces(tCOMBAT[a][i].GX, tCOMBAT[a][i].GY)
        if next(spaces)~=nil then
          j = math.random(#spaces)
          table.insert(tCOMBAT[a], deepcopy(tCOMBAT[a][i]))
          tCOMBAT[a][#tCOMBAT[a]].Wounds = 0
          tCOMBAT[a][#tCOMBAT[a]].GX, tCOMBAT[a][#tCOMBAT[a]].GY = spaces[j].X, spaces[j].Y
          tCOMBAT[a][#tCOMBAT[a]].X, tCOMBAT[a][#tCOMBAT[a]].Y = gridtoscreen(spaces[j].X, spaces[j].Y)
          addfloatytext(tCOMBAT[a][i], "+Breeds+")
        end
      end
      if gethasability(tCOMBAT[a][i], "Forgetful") and math.random(100)<=15 and #tHAND[a]<tSETTINGS.MaxCardsHand then
        addfloatytext(tCOMBAT[a][i], "-Forgetful-")
        sendcard("Hand", tCOMBAT[a][i], "Combat", a, i)
        removed = true
      end
      if removed == false and tCOMBAT[a][i].Effect[11]>0 then
        addfloatytext(tCOMBAT[a][i], "-Poison-")
        if gethasability(tCOMBAT[a][i], "Poison Weakness") then removed = damage(tCOMBAT[a][i], "Combat", a, i, 2+getarmour(tCOMBAT[a][i]))
        else removed = damage(tCOMBAT[a][i], "Combat", a, i, 1+getarmour(tCOMBAT[a][i])) end
      end
      if removed == false and tCOMBAT[a][i].Effect[12]>0 then
        addfloatytext(tCOMBAT[a][i], "-Poison-")
        if gethasability(tCOMBAT[a][i], "Poison Weakness") then removed = damage(tCOMBAT[a][i], "Combat", a, i, 6+getarmour(tCOMBAT[a][i]))
        else removed = damage(tCOMBAT[a][i], "Combat", a, i, 3+getarmour(tCOMBAT[a][i])) end
      end
      if removed == false and tCOMBAT[a][i].Effect[13]>0 then
        addfloatytext(tCOMBAT[a][i], "-Poison-")
        if gethasability(tCOMBAT[a][i], "Poison Weakness") then removed = damage(tCOMBAT[a][i], "Combat", a, i, 10+getarmour(tCOMBAT[a][i]))
        else removed = damage(tCOMBAT[a][i], "Combat", a, i, 5+getarmour(tCOMBAT[a][i])) end
      end
      if removed == false and (tCOMBAT[a][i].Effect[1]>0 or tCOMBAT[a][i].Effect[3]>0) then tCOMBAT[a][i].Steps = tCOMBAT[a][i].Steps+getreach(tCOMBAT[a][i]) end
    end
  end
  if next(tCOMBAT[b]) ~= nil then
    for i = #tCOMBAT[b], 1, -1 do
      tCOMBAT[b][i].FaceUp = true
      tCOMBAT[b][i].Clickable = true
      tCOMBAT[b][i].Steps = 0
      for j = 1, #tEFFECTS, 1 do
        if tCOMBAT[b][i].Effect[j]<100 then tCOMBAT[b][i].Effect[j] = math.max(tCOMBAT[b][i].Effect[j]-1, 0) end
      end
      if gethasability(tCOMBAT[b][i], "Nocturnal") then
        if gDAYNIGHT>=6 then tCOMBAT[b][i].Effect[10] = 1
        else tCOMBAT[b][i].Effect[9] = 1 end
      end
      if tCOMBAT[b][i].Effect[1]>0 or tCOMBAT[b][i].Effect[3]>0 then tCOMBAT[b][i].Steps = tCOMBAT[b][i].Steps+getreach(tCOMBAT[b][i]) end
    end
  end
  if next(tHAND[a]) ~= nil then
    for i = #tHAND[a], 1, -1 do
      if tCONTROL[a]=="Player" then 
        tHAND[a][i].FaceUp = true
        tHAND[a][i].Clickable = true
      end
      for j = 1, #tEFFECTS, 1 do
        if tHAND[a][i].Effect[j]<100 then tHAND[a][i].Effect[j] = math.max(tHAND[a][i].Effect[j]-1, 0) end
      end
    end
  end
  if next(tHAND[b]) ~= nil then
    for i = #tHAND[b], 1, -1 do
      tHAND[b][i].FaceUp = false
      tHAND[b][i].Clickable = false
      for j = 1, #tEFFECTS, 1 do
        if tHAND[b][i].Effect[j]<100 then tHAND[b][i].Effect[j] = math.max(tHAND[b][i].Effect[j]-1, 0) end
      end
    end
  end
  if next(tGRAVE[a]) ~= nil then
    for i = #tGRAVE[a], 1, -1 do
      tGRAVE[a][i].FaceUp = true
      tGRAVE[a][i].Clickable = false
      for j = 1, #tEFFECTS, 1 do
        if tGRAVE[a][i].Effect[j]<100 then tGRAVE[a][i].Effect[j] = math.max(tGRAVE[a][i].Effect[j]-1, 0) end
      end
    end
    tGRAVE[a][#tGRAVE[a]].Clickable = true
  end
  if next(tGRAVE[b]) ~= nil then
    for i = #tGRAVE[b], 1, -1 do
      tGRAVE[b][i].FaceUp = true
      tGRAVE[b][i].Clickable = false
      for j = 1, #tEFFECTS, 1 do
        if tGRAVE[b][i].Effect[j]<100 then tGRAVE[b][i].Effect[j] = math.max(tGRAVE[b][i].Effect[j]-1, 0) end
      end
    end
    tGRAVE[b][#tGRAVE[b]].Clickable = true
  end
  if tCONTROL[a] == "Player" and tCONTROL[b] == "Player" then gGAMEMODE = "newturn"
  else
    gGAMEMODE = "game"
    if tCONTROL[a] == "AI" then
      i = AIscore(a)
      AIlastmove = love.timer.getTime()
      --ailog("INITIAL BOARDSCORE: "..i)
      AIbest_score = i
      AIbest_start = {}
      AIloop = 0
    end
  end
  tCLICK = {}
  gZOOMED = false
  gSACRIFICED = false
  gACTION = ""
  AIsimulation=false
  arrangehand(pl)
end

function calculatemoves(pl)  -- what moves can player PL make?
  local moves = {false, false, false, false, false, false, false, false, false} 
  local source = tCLICK[1].Source
  local card = {}
  local i
  if source == "Hand" then card = tHAND[tCLICK[1].Player][tCLICK[1].Index]
  elseif source == "Combat" then card = tCOMBAT[tCLICK[1].Player][tCLICK[1].Index]
  elseif source == "Grave" then card = tGRAVE[tCLICK[1].Player][tCLICK[1].Index]
  elseif source == "Deck" then card = tDECK[tCLICK[1].Player][tCLICK[1].Index]
  else end
  -- 1=March (play), 2=Action (play), 3=Move/Attack, 4=Target, 5=Spec1, 6=Spec2, 7=Spec3, 8=Spec4, 9=Sacrifice
  if next(card)~=nil then
    if card.Target==true and gACTION~="" then moves[4] = true
    elseif source=="Hand" and gTURN==pl and gACTION=="" and tPOINTS[pl]>=card.Tier and (card.Type=="Ally" or card.Type=="Hero") and getishomefree(pl) then moves[1] = true
    elseif source=="Hand" and gTURN==pl and gACTION=="" and tPOINTS[pl]>=card.Tier and card.Type=="Action" then moves[2] = true
    elseif source=="Combat" and gTURN==pl and gACTION=="" and getreach(card)>0  then moves[3]= true
    else end
    if source=="Combat" and gTURN==pl and gACTION=="" then
      for i = 1, 4, 1 do
        if card["Spec"..i]~=nil and card["Charge"..i]>0 and card["Charge"..i]==getrechargetime(card["Spec"..i]) then moves[4+i] = true end
      end
    end
    if source=="Hand" and gTURN==pl and gACTION=="" and gSACRIFICED==false then moves[9] = true end
  end
  return moves
end

function findtargets()  -- which cards can be targetted for this action?
  local a,b = gTURN, 3-gTURN
  local reach = 0
  local found = false
  local origin
  if next(tCLICK)~=nil then 
    if tCLICK[#tCLICK].Source == "Hand" then origin = tHAND[tCLICK[#tCLICK].Player][tCLICK[#tCLICK].Index]
    elseif tCLICK[#tCLICK].Source == "Combat" then origin = tCOMBAT[tCLICK[#tCLICK].Player][tCLICK[#tCLICK].Index]
    elseif tCLICK[#tCLICK].Source == "Grave" then origin = tGRAVE[tCLICK[#tCLICK].Player][tCLICK[#tCLICK].Index]
    elseif tCLICK[#tCLICK].Source == "Deck" then origin = tDECK[tCLICK[#tCLICK].Player][tCLICK[#tCLICK].Index]
    else end
    reach = getreach(origin)
  end
  if gACTION ~= "" then
    if next(tDECK[a]) ~= nil then
      for i = 1, #tDECK[a], 1 do 
        tDECK[a][i].Target = false
      end
    end
    if next(tHAND[a]) ~= nil then
      for i = 1, #tHAND[a], 1 do
        tHAND[a][i].Target = false
      end
    end
    if next(tCOMBAT[a]) ~= nil then
      for i = 1, #tCOMBAT[a], 1 do
        tCOMBAT[a][i].Target = false
        if gACTION=="Cage" and tCOMBAT[a][i].Size~="Tiny" then
          tCOMBAT[a][i].Target = true
          found = true
        elseif gACTION=="Calm Mammal" and tCOMBAT[a][i].Effect[4]>0 then
          tCOMBAT[a][i].Target = true
          found = true
        elseif gACTION=="Cure Mild Poison" and (tCOMBAT[a][i].Effect[11]>0 or tCOMBAT[a][i].Effect[12]>0 or tCOMBAT[a][i].Effect[13]>0) then
          tCOMBAT[a][i].Target = true
          found = true
        elseif gACTION=="Spawn Ant" and string.find(tCOMBAT[a][i].Name, " Ant") and tCOMBAT[a][i].Name~="Queen Ant" then
          tCOMBAT[a][i].Target = true
          found = true
        elseif gACTION=="Sprint" and getreach(tCOMBAT[a][i])>0 then
          tCOMBAT[a][i].Target = true
          found = true
        elseif gACTION=="Thorns" then
          tCOMBAT[a][i].Target = true
          found = true
        else end
      end
    end
    if next(tGRAVE[a]) ~= nil then
      for i = 1, #tGRAVE[a], 1 do 
        tGRAVE[a][i].Target = false
      end
    end
    if next(tDECK[b]) ~= nil then
      for i = 1, #tDECK[b], 1 do 
        tDECK[b][i].Target = false
      end
    end
    if next(tHAND[b]) ~= nil then
      for i = 1, #tHAND[b], 1 do 
        tHAND[b][i].Target = false
        if gACTION=="Snatch Card" and #tHAND[a]<tSETTINGS.MaxCardsHand then 
          tHAND[b][i].Target = true
          found = true
        end
      end
    end
    if next(tCOMBAT[b]) ~= nil then
      for i = 1, #tCOMBAT[b], 1 do
        tCOMBAT[b][i].Target = false
        if gACTION=="Move/Attack" and tHEX[tCOMBAT[b][i].GY][tCOMBAT[b][i].GX].Moves<=reach and tCOMBAT[b][i].Effect[2]==0 and gCEASEFIRE==0 then 
          tCOMBAT[b][i].Target = true
          found = true
        elseif gACTION=="Cage" and tCOMBAT[b][i].Size~="Tiny" and tCOMBAT[b][i].Effect[2]==0 then
          tCOMBAT[b][i].Target = true
          found = true
        elseif gACTION=="Charm Mammal" and tCOMBAT[b][i].Class=="Mammal" and getisadjacent(origin, tCOMBAT[b][i]) then
          tCOMBAT[b][i].Target = true
          found = true
        elseif gACTION=="Kitten Scratch" and tCOMBAT[b][i].Effect[2]==0 then
          tCOMBAT[b][i].Target = true
          found = true
        elseif gACTION=="Mesmerise" and getisadjacent(origin, tCOMBAT[b][i]) then
          tCOMBAT[b][i].Target = true
          found = true
        elseif gACTION=="Paralysis (Minor)" and getisadjacent(origin, tCOMBAT[b][i]) and (tCOMBAT[b][i].Size=="Tiny" or tCOMBAT[b][i].Size=="Small") and tCOMBAT[b][i].Effect[2]==0 then
          tCOMBAT[b][i].Target = true
          found = true
        elseif gACTION=="Poison (Minor)" and getisadjacent(origin, tCOMBAT[b][i]) and tCOMBAT[b][i].Effect[2]==0 then
          tCOMBAT[b][i].Target = true
          found = true
        elseif gACTION=="Snake Coil" and (tCOMBAT[b][i].Size=="Small" or tCOMBAT[b][i].Size=="Medium") and getisadjacent(origin, tCOMBAT[b][i]) and tCOMBAT[b][i].Effect[2]==0 then
          tCOMBAT[b][i].Target = true
          found = true
        elseif gACTION=="Thorns" and tCOMBAT[b][i].Effect[2]==0 then
          tCOMBAT[b][i].Target = true
          found = true
        elseif gACTION=="Web (Small)" and getisadjacent(origin, tCOMBAT[b][i]) and tCOMBAT[b][i].Size=="Tiny" and tCOMBAT[b][i].Effect[2]==0 then
          tCOMBAT[b][i].Target = true
          found = true
        else end
      end
    end
    if next(tGRAVE[b]) ~= nil then
      for i = 1, #tGRAVE[b], 1 do
        tGRAVE[b][i].Target = false
      end
    end
  end
  return found
end

function cleartargets()
  local i, j
  for i = 1, 2, 1 do
    if next(tHAND[i]) ~= nil then
      for j = 1, #tHAND[i], 1 do tHAND[i][j].Target = false end
    end
    if next(tDECK[i]) ~= nil then
      for j = 1, #tDECK[i], 1 do tDECK[i][j].Target = false end
    end
    if next(tCOMBAT[i]) ~= nil then
      for j = 1, #tCOMBAT[i], 1 do tCOMBAT[i][j].Target = false end
    end
    if next(tGRAVE[i]) ~= nil then
      for j = 1, #tGRAVE[i], 1 do tGRAVE[i][j].Target = false end
    end
  end
end

function sendcard(dest, card, loc, pl, index)   -- send card to destination
  local a, b = pl, 3-pl
  local i, gx, gy
  if dest == "Combat" then
    if pl == 1 then gy = gGRID_HEIGHT
    else gy = 1 end
    repeat
      i = math.random(5)
    until tHEX[gy][i].Player == 0
    gx = i
  end
  card.Animate = true
  card.AniCount = tSETTINGS.AniTime
  card.OriginalX, card.OriginalY = card.X, card.Y
  card.Hover = false
  card.Steps = 0
  for i = 1, #tEFFECTS, 1 do card.Effect[i] = 0 end
  for i = 1, 4, 1 do card["Charge"..i] = 0 end
  if dest == "Hand" then 
    card.TargetX, card.TargetY = xHAND[pl], yHAND[pl]
    if tCONTROL[pl]=="Player" and gTURN==pl then card.FaceUp = true
    else card.FaceUp = false end
    if gTURN == pl then card.Clickable = true
    else card.Clickable = false end
    table.insert(tHAND[pl], deepcopy(card))
  elseif dest == "Grave" then 
    card.TargetX, card.TargetY = xGRAVE[pl], yGRAVE[pl]
    card.FaceUp = true
    card.Angle = 11 - math.random(20)
    if next(tGRAVE[pl])~=nil then tGRAVE[pl][#tGRAVE[pl]].Clickable = false end
    card.Clickable = true
    table.insert(tGRAVE[pl], deepcopy(card))
  elseif dest == "Deck" then 
    card.TargetX, card.TargetY = xDECK[pl], yDECK[pl]
    card.FaceUp = false
    card.Angle = 0
    card.Clickable = false
    if next(tDECK[pl])~=nil then table.insert(tDECK[pl], math.random(#tDECK[pl]), deepcopy(card))
    else table.insert(tDECK[pl], deepcopy(card)) end
  elseif dest == "Combat" then
    card.TargetX, card.TargetY = gridtoscreen(gx, gy)
    card.GX, card.GY = gx, gy
    card.FaceUp = true
    card.Angle = 0
    card.Clickable = true
    card.Steps = card.Steps + getreach(card)
    tHEX[gy][gx].Player = pl
    if gethasability(card, "Nocturnal") then
      if gDAYNIGHT>=6 then card.Effect[10] = 1
      else card.Effect[9] = 1 end
    end
    if gethasability(card, "Defend the Queen") then
      if next(tCOMBAT[a])~=nil then
        for i = 1, #tCOMBAT[a], 1 do
          if tCOMBAT[a][i].Name=="Queen Ant" and string.find(card.Name, "Ant") then 
            card.Effect[8]=100
            addfloatytext(card, "+Defender+")
          elseif tCOMBAT[a][i].Name=="Queen Bee" and string.find(card.Name, "Bee") then 
            card.Effect[8]=100
            addfloatytext(card, "+Defender+")
          else end
        end
      end
    elseif string.sub(card.Name, 1, 5)=="Queen" then
      if next(tCOMBAT[a])~=nil then
        for i = 1, #tCOMBAT[a], 1 do
          if gethasability(tCOMBAT[a][i], "Defend the Queen") then
            if card.Name=="Queen Ant" and string.find(tCOMBAT[a][i].Name, "Ant") then 
              tCOMBAT[a][i].Effect[8]=100
              addfloatytext(tCOMBAT[a][i], "+Defender+")
            elseif card.Name=="Queen Bee" and string.find(tCOMBAT[a][i].Name, "Bee") then 
              tCOMBAT[a][i].Effect[8]=100
              addfloatytext(tCOMBAT[a][i], "+Defender+")
            else end
          end
        end
      end
    else end
    if gethasability(card, "Scary to Mammals") then
      if next(tCOMBAT[b])~=nil then
        for i = 1, #tCOMBAT[b], 1 do
          if tCOMBAT[b][i].Class=="Mammal" and (tCOMBAT[b][i].Size=="Medium" or tCOMBAT[b][i].Size=="Small" or tCOMBAT[b][i].Size=="Tiny") then
            if math.random(100)<=50 or gethasability(tCOMBAT[b][i], "Fearful") then
              addfloatytext(tCOMBAT[b][i], "-Fear-")
              tCOMBAT[b][i].Effect[4] = math.max(tCOMBAT[b][i].Effect[4], 2)
            else addfloatytext(tCOMBAT[b][i], "+Resist Fear+") end
          end
        end
      end
    elseif card.Class=="Mammal" and (card.Size=="Medium" or card.Size=="Small" or card.Size=="Tiny") then
      if next(tCOMBAT[b])~=nil then
        for i = 1, #tCOMBAT[b], 1 do
          if gethasability(tCOMBAT[b][i], "Scary to Mammals") then
            if math.random(100)<=50 or gethasability(card, "Fearful") then
              addfloatytext(card, "-Fear-")
              card.Effect[4] = math.max(card.Effect[4], 2)
            else addfloatytext(card, "+Resist Fear+") end
          end
        end
      end
    else end
    if gethasability(card, "Excavate") and #tHAND[a]<tSETTINGS.MaxCardsHand then
      j = math.min(tSETTINGS.MaxCardsHand - #tHAND[a], 3)
      for i = 1, j, 1 do
        dealcard(a)
      end
    end
    if gethasability(card, "Industrious") then
      tPOINTS[a] = math.min(tPOINTS[a]+1, tBUDGET[a])
      addfloatytext(card, "+Industrious+")
    end
    if gethasability(card, "Stealthy") then
      card.Effect[2] = math.max(card.Effect[2], 100)
      addfloatytext(card, "+Stealthy+")
    end
    if gethasability(card, "Surprise Attack") then
      card.Steps = 0
      addfloatytext(card, "+Surprise Attack+")
    end
    table.insert(tCOMBAT[pl], deepcopy(card))
  else end
  if loc == "Hand" then
    table.remove(tHAND[pl], index)
    arrangehand(pl)
  elseif loc == "Combat" then
    tHEX[card.GY][card.GX].Player = 0
    table.remove(tCOMBAT[pl], index)
  elseif loc == "Deck" then table.remove(tDECK[pl], index)
  elseif loc == "Grave" then
    table.remove(tGRAVE[pl], index)
    if next(tGRAVE[pl])~=nil then tGRAVE[pl][#tGRAVE[pl]].Clickable = true end
  else end
  if dest == "Hand" then arrangehand(pl) end
end

function playaction(card, pl)  -- Play action or use special ability
  local r = false              -- Flag that playing the action was successful
  local a, b = pl, 3-pl        -- Friendly & foe
  local targets = {}           -- List of potential targets to randomly pick from
  local x, y, d, bx, by, bd    -- Used in revive bees to find closest non-occupied point
  local i, j
  if getabilitytargets(gACTION) > 0 then       -- These actions have a target, same step for each
    if findtargets() == true then
      if getabilitytargets(gACTION) > 1 then addpopup("Choose targets")
      else addpopup("Choose target") end
      r = true
    else
      addpopup("No target available")
      gACTION = ""
    end
  else                                   -- Now alphabetical, these actions have no target, do them!
    if gACTION == "Charge" then
      addpopup("Choose location")
      refreshmoves()
      r = true
    elseif gACTION == "Dash" then
      addpopup(card.Name.." makes a dash for it")
      card.Effect[15] = math.max(card.Effect[15], 2)
      r = true
    elseif gACTION == "Day to Night" then
      addpopup("P"..a.." calls forth the moon")
      r = true
      gDAYNIGHT = 1
      for j = 1, 2, 1 do
        if next(tCOMBAT[j])~=nil then
          for i = 1, #tCOMBAT[j], 1 do
            if gethasability(tCOMBAT[j][i], "Nocturnal") then tCOMBAT[j][i].Effect[9] = 1 end
          end
        end
      end
    elseif gACTION == "Feed Birds" then
      addpopup("P"..a.." feeds the birds")
      if next(tCOMBAT[a])~=nil then
        for i = 1, #tCOMBAT[a], 1 do
          if tCOMBAT[a][i].Class == "Bird" then
            addfloatytext(tCOMBAT[a][i], "+Fed+")
            addfloatytext(tCOMBAT[a][i], "+"..math.min(tCOMBAT[a][i].Wounds,10).." Health")
            tCOMBAT[a][i].Wounds = math.max(tCOMBAT[a][i].Wounds-10, 0)
            tCOMBAT[a][i].Effect[5] = 100
            r = true
          end
        end
      end
    elseif gACTION == "Feed Fish" then
      addpopup("P"..a.." feeds the fish")
      if next(tCOMBAT[a])~=nil then
        for i = 1, #tCOMBAT[a], 1 do
          if tCOMBAT[a][i].Class == "Fish" then
            addfloatytext(tCOMBAT[a][i], "+Fed+")
            addfloatytext(tCOMBAT[a][i], "+"..math.min(tCOMBAT[a][i].Wounds,10).." Health")
            tCOMBAT[a][i].Wounds = math.max(tCOMBAT[a][i].Wounds-10, 0)
            tCOMBAT[a][i].Effect[5] = 100
            r = true
          end
        end
      end
    elseif gACTION == "Feed Insects" then
      addpopup("P"..a.." feeds the insects")
      if next(tCOMBAT[a])~=nil then
        for i = 1, #tCOMBAT[a], 1 do
          if tCOMBAT[a][i].Class == "Insect" then
            addfloatytext(tCOMBAT[a][i], "+Fed+")
            addfloatytext(tCOMBAT[a][i], "+"..math.min(tCOMBAT[a][i].Wounds,10).." Health")
            tCOMBAT[a][i].Wounds = math.max(tCOMBAT[a][i].Wounds-10, 0)
            tCOMBAT[a][i].Effect[5] = 100
            r = true
          end
        end
      end
    elseif gACTION == "Feed Mammals" then
      addpopup("P"..a.." feeds the mammals")
      if next(tCOMBAT[a])~=nil then
        for i = 1, #tCOMBAT[a], 1 do
          if tCOMBAT[a][i].Class == "Mammal" then
            addfloatytext(tCOMBAT[a][i], "+Fed+")
            addfloatytext(tCOMBAT[a][i], "+"..math.min(tCOMBAT[a][i].Wounds,10).." Health")
            tCOMBAT[a][i].Wounds = math.max(tCOMBAT[a][i].Wounds-10, 0)
            tCOMBAT[a][i].Effect[5] = 100
            r = true
          end
        end
      end
    elseif gACTION == "Hide Small Creatures" then
      addpopup("The small creatures are hiding!")
      r = true
      for j = 1, 2, 1 do
        if next(tCOMBAT[j])~=nil then
          for i = 1, #tCOMBAT[j], 1 do
            if tCOMBAT[j][i].Size=="Tiny" or tCOMBAT[j][i].Size=="Small" then
              tCOMBAT[j][i].Effect[2] = math.max(tCOMBAT[j][i].Effect[2], 2)
              addfloatytext(tCOMBAT[j][i], "+Hidden+")
            end
          end
        end
      end
    elseif gACTION == "High Winds" then
      addpopup("A blustery gale!")
      r = true
      for j = 1, 2, 1 do
        if next(tCOMBAT[j])~=nil then
          for i = #tCOMBAT[j], 1, -1 do
            if tCOMBAT[j][i].Size=="Tiny" then
              addfloatytext(tCOMBAT[j][i], "-Blown away-")
              sendcard("Grave", tCOMBAT[j][i], "Combat", j, i)
            end
          end
        end
      end
    elseif gACTION == "Hoover" then
      addpopup("P"..a.." uses the hoover")
      r = true
      for j = 1, 2, 1 do
        if next(tCOMBAT[j])~=nil then
          for i = #tCOMBAT[j], 1, -1 do
            if tCOMBAT[j][i].Size=="Tiny" and tCOMBAT[j][i].Class~="Fish" then
              addfloatytext(tCOMBAT[j][i], "-Hoovered-")
              sendcard("Grave", tCOMBAT[j][i], "Combat", j, i)
            elseif (tCOMBAT[j][i].Size=="Small" or tCOMBAT[j][i].Size=="Medium") and tCOMBAT[j][i].Class~="Fish" then
              addfloatytext(tCOMBAT[j][i], "-Fear-")
              tCOMBAT[j][i].Effect[4] = math.max(tCOMBAT[j][i].Effect[4], 2)
            else end
          end
        end
      end
    elseif gACTION == "Insecticide" then
      addpopup("P"..a.." sprays insecticide")
      for j = 1, 2, 1 do
        if next(tCOMBAT[j])~=nil then
          for i = #tCOMBAT[j], 1, -1 do
            if tCOMBAT[j][i].Class=="Insect" then
              addfloatytext(tCOMBAT[j][i], "-Eliminated-")
              sendcard("Grave", tCOMBAT[j][i], "Combat", j, i)
              r = true
            end
          end
        end
      end
    elseif gACTION == "Intimidate" then
      addpopup(card.Name.." intimidates")
      r = true
      if next(tCOMBAT[b])~=nil then
        for i = 1, #tCOMBAT[b], 1 do
          if tCOMBAT[b][i].Class=="Mammal" and (tCOMBAT[b][i].Size=="Medium" or tCOMBAT[b][i].Size=="Small" or tCOMBAT[b][i].Size=="Tiny") and getisadjacent(card, tCOMBAT[b][i]) then
            addfloatytext(tCOMBAT[b][i], "-Fear-")
            tCOMBAT[b][i].Effect[4] = math.max(tCOMBAT[b][i].Effect[4], 2)
          end
        end
      end
    elseif gACTION == "Mow Lawn" then
      addpopup("P"..a.." mows the lawn")
      r = true
      if next(tCOMBAT[b])~=nil then
        for i = #tCOMBAT[b], 1, -1 do
          if tCOMBAT[b][i].Size~="Tiny" and tCOMBAT[b][i].Class~="Fish" and tCOMBAT[b][i].Class~="Bird" then
            addfloatytext(tCOMBAT[b][i], "-Mown-")
            damage(tCOMBAT[b][i], "Combat", b, i, 5)
          end
        end
      end
    elseif gACTION == "Night to Day" then
      addpopup("P"..a.." wakes the sun")
      r = true
      gDAYNIGHT = 6
      for j = 1, 2, 1 do
        if next(tCOMBAT[j])~=nil then
          for i = 1, #tCOMBAT[j], 1 do
            if gethasability(tCOMBAT[j][i], "Nocturnal") then tCOMBAT[j][i].Effect[10] = 1 end
          end
        end
      end
    elseif gACTION == "Night Vision" then
      addpopup(card.Name.." sees everything")
      if next(tCOMBAT[b])~=nil then
        for i = 1, #tCOMBAT[b], 1 do
          if tCOMBAT[b][i].Effect[2]>0 then table.insert(targets, i) end
        end
      end
      if next(targets)~=nil then
        r = true
        i = targets[math.random(#targets)]
        tCOMBAT[b][i].Effect[2]=0
        addfloatytext(tCOMBAT[b][i], "-Revealed-")
      end
    elseif gACTION == "No Hiding" then
      addpopup("There's no hiding from "..card.Name)
      if next(tCOMBAT[b])~=nil then
        for i = 1, #tCOMBAT[b], 1 do
          if tCOMBAT[b][i].Effect[2]>0 then
            r = true
            tCOMBAT[b][i].Effect[2]=0
            addfloatytext(tCOMBAT[b][i], "-Revealed-")
          end
        end
      end
    elseif gACTION == "Rage" then
      addpopup(card.Name.." is enraged")
      addfloatytext(card, "+Rage+")
      r = true
      card.Effect[14] = math.max(card.Effect[14], 1)
    elseif gACTION == "Rain Delay" then
      addpopup("A torrential downpour")
      gCEASEFIRE = math.max(gCEASEFIRE, 3)
      r = true
    elseif gACTION == "Reinforcements" then
      addpopup("P"..a.." calls in reinforcements")
      if #tHAND[a]<tSETTINGS.MaxCardsHand+1 then   -- Allow one extra because we discard the action card
        r = true
        dealcard(a)
        if #tHAND[a]<tSETTINGS.MaxCardsHand+1 then dealcard(a) end
      end
    elseif gACTION == "Repel Insects" then
      addpopup("P"..a.." sprays insect repellent")
      for j = 1, 2, 1 do
        if next(tCOMBAT[j])~=nil then
          for i = #tCOMBAT[j], 1, -1 do
            if tCOMBAT[j][i].Class == "Insect" then
              r = true
              if math.random(100)<=75 then 
                addfloatytext(tCOMBAT[j][i], "-Repelled-")
                sendcard("Deck", tCOMBAT[j][i], "Combat", j, i)
              end
            end
          end
        end
      end
    elseif gACTION == "Revive Bees" then
      if next(tGRAVE[a])~=nil then
        addpopup("P"..a.." revives the bees")
        r = true
        for i = #tGRAVE[a], 1, -1 do
          if tGRAVE[a][i].Name=="Drone Bee" then
            bx, by, bd = 0, 0, math.huge
            for y = 1, gGRID_HEIGHT, 1 do
              for x = 1, gGRID_WIDTH, 1 do
                if tHEX[y][x].Player==0 then
                  if a==1 then d = getflatdist(x, y, 5, 9)
                  else d = getflatdist(x, y, 5, 1) end
                  if d<bd then bx, by, bd = x, y, d end
                end
              end
            end
            if bx~=0 and by~=0 then
              tGRAVE[a][i].Animate = true
              tGRAVE[a][i].AniCount = tSETTINGS.AniTime
              tGRAVE[a][i].OriginalX, tGRAVE[a][i].OriginalY = tGRAVE[a][i].X, tGRAVE[a][i].Y
              tGRAVE[a][i].Steps = tGRAVE[a][i].Steps + getreach(tGRAVE[a][i])
              for j = 1, 4, 1 do
                tGRAVE[a][i]["Charge"..j]=0
              end
              tGRAVE[a][i].Wounds = 0
              tGRAVE[a][i].FaceUp = true
              tGRAVE[a][i].Clickable = true
              tGRAVE[a][i].GX, tGRAVE[a][i].GY = bx, by
              tHEX[by][bx].Player = a
              tGRAVE[a][i].TargetX, tGRAVE[a][i].TargetY = gridtoscreen(bx, by)
              table.insert(tCOMBAT[a], deepcopy(tGRAVE[a][i]))
              table.remove(tGRAVE[a], i)
            end
          end
        end
      end
    elseif gACTION == "Reveal Cards" then
      if next(tHAND[b])~=nil then
        addpopup("P"..a.." peeks at P"..b.."'s cards")
        r = true
        for i = 1, #tHAND[b], 1 do
          tHAND[b][i].FaceUp = true
        end
      end
    elseif gACTION == "Resourceful" then
      addpopup(card.Name.." is very resourceful")
      if #tHAND[a]<tSETTINGS.MaxCardsHand then
        addfloatytext(card, "+Extra Card+")
        r = true
        dealcard(a)
      end
    elseif gACTION == "Scratch" then
      addpopup(card.Name.." swipes at those around it")
      if next(tCOMBAT[b])~=nil then
        for i = #tCOMBAT[b], 1, -1 do
          if getisadjacent(card, tCOMBAT[b][i]) and tCOMBAT[b][i].Effect[2]==0 then 
            addfloatytext(tCOMBAT[b][i], "-Scratched-")
            damage(tCOMBAT[b][i], "Combat", b, i, 1)
            r = true
          end
        end
      end
    elseif gACTION == "Shine Light" and gDAYNIGHT<=5 then
      r = true
      addpopup("P"..a.." shines a torch")
      for j = 1, 2, 1 do
        if next(tCOMBAT[j])~=nil then
          for i = 1, #tCOMBAT[j], 1 do
            if tCOMBAT[j][i].Class=="Insect" then
              addfloatytext(tCOMBAT[j][i], "+Light+")
              tCOMBAT[j][i].Effect[6] = math.max(tCOMBAT[j][i].Effect[6], 4)
            elseif gethasability(tCOMBAT[j][i], "Nocturnal") then
              addfloatytext(tCOMBAT[j][i], "-Blinded-")
              tCOMBAT[j][i].Effect[7] = math.max(tCOMBAT[j][i].Effect[7], 4)
            else end
          end
        end
      end
    elseif gACTION == "Spray Water" then
      addpopup("P"..a.." sprays water")
      r = true
      for j = 1, 2, 1 do
        if next(tCOMBAT[j])~=nil then
          for i = #tCOMBAT[j], 1, -1 do
            addfloatytext(tCOMBAT[j][i], "-Drenched-")
            if tCOMBAT[j][i].Class == "Insect" then damage(tCOMBAT[j][i], "Combat", j, i, 2)
            elseif tCOMBAT[j][i].Class=="Fish" or tCOMBAT[j][i].Class=="Annelid" or tCOMBAT[j][i].Class=="Mollusc" or tCOMBAT[j][i].Class=="Reptile" or tCOMBAT[j][i].Class=="Crustacean" then tCOMBAT[j][i].Effect[16] = math.max(tCOMBAT[j][i].Effect[16], 4)
            else tCOMBAT[j][i].Effect[17] = math.max(tCOMBAT[j][i].Effect[17], 4) end
          end
        end
      end
    elseif gACTION == "Swarm" then
      addpopup("P"..a.." calls in a swarm of bees")
      r = true
      if next(tCOMBAT[b])~=nil then
        for i = #tCOMBAT[b], 1, -1 do
          addfloatytext(tCOMBAT[b][i], "-Swarm-")
          damage(tCOMBAT[b][i], "Combat", b, i, 5)
        end
      end
      if next(tCOMBAT[a])~=nil then
        for i = #tCOMBAT[a], 1, -1 do
          if string.find(tCOMBAT[a][i].Name, " Bee") then 
            addfloatytext(tCOMBAT[a][i], "-Swarm-")
            sendcard("Grave", tCOMBAT[a][i], "Combat", a, i)
          end
        end
      end
    elseif gACTION == "Tail Whip" then
      addpopup(card.Name.." strikes with its tail")
      r = true
      addfloatytext(card, "+Tail Whip+")
      if next(tCOMBAT[b])~=nil then
        for i = #tCOMBAT[b], 1, -1 do
          if getisadjacent(card, tCOMBAT[b][i]) then
            tCOMBAT[b][i].Effect[2] = 0
            addfloatytext(tCOMBAT[b][i], "-Whipped-")
            damage(tCOMBAT[b][i], "Combat", b, i, 2)
          end
        end
      end
    elseif gACTION == "Trip to the Vet" then
      r = true
      addpopup("Oh no! It's the veterinarian!")
      for j = 1, 2, 1 do
        if next(tCOMBAT[j])~=nil then
          for i = #tCOMBAT[j], 1, -1 do
            if tCOMBAT[j][i].Size=="Small" or tCOMBAT[j][i].Size=="Medium" and math.random(100)<=50 then
              tCOMBAT[j][i].Wounds = 0
              if #tHAND[j]<tSETTINGS.MaxCardsHand then sendcard("Hand", tCOMBAT[j][i], "Combat", j, i)
              else sendcard("Deck", tCOMBAT[j][i], "Combat", j, i) end
            end
          end
        end
      end
    elseif gACTION == "Webs" then
      addpopup("P"..a.." spins some webs")
      r = true
      if next(tCOMBAT[a])~=nil then
        for i = 1, #tCOMBAT[a], 1 do
          if tCOMBAT[a][i].Size=="Tiny" then
            addfloatytext(tCOMBAT[a][i], "-Webbed-")
            tCOMBAT[a][i].Effect[3] = math.max(tCOMBAT[a][i].Effect[3], 4)
          end
        end
      end
    elseif gACTION == "Wild Flowers" then
      addpopup("P"..a.." brought flowers")
      for j = 1, 2, 1 do
        if next(tCOMBAT[j])~=nil then
          for i = 1, #tCOMBAT[j], 1 do
            if tCOMBAT[j][i].Class=="Insect" then
              r = true
              addfloatytext(tCOMBAT[j][i], "-Distracted-")
              tCOMBAT[j][i].Effect[4] = math.max(tCOMBAT[j][i].Effect[4], 2)
            end
          end
        end
      end
    elseif gACTION == "Yip! Yip!" then
      addpopup(card.Name.." barks")
      r = true
      if next(tCOMBAT[b])~=nil then
        for i = 1, #tCOMBAT[b], 1 do
          if tCOMBAT[b][i].Size=="Tiny" or tCOMBAT[b][i].Size=="Small" then
            if math.random(100)<=50 then
              addfloatytext(tCOMBAT[b][i], "-Fear-")
              tCOMBAT[b][i].Effect[4] = math.max(tCOMBAT[b][i].Effect[4], 2)
            else addfloatytext(tCOMBAT[b][i], "+Resisted+") end
          end
        end
      end
    else end
  end
  return r
end

function domove(x, y)  -- move the card to this space
  local pl = tCLICK[#tCLICK].Player
  local card = tCOMBAT[pl][tCLICK[#tCLICK].Index]
  local i
  card.OriginalX, card.OriginalY = card.X, card.Y
  card.TargetX, card.TargetY = gridtoscreen(x, y)
  card.Animate = true
  card.AniCount = tSETTINGS.AniTime
  tHEX[card.GY][card.GX].Player=0
  card.GX, card.GY = x, y
  tHEX[card.GY][card.GX].Player=tCLICK[#tCLICK].Player
  if gethasability(card, "Energetic") and math.random(100)<=20 then addfloatytext(card, "+Energetic+")
  else card.Steps = card.Steps + tHEX[y][x].Moves end
  for i = 1, 4, 1 do
    if card["Spec"..i]~=nil and card["Spec"..i]=="Charge" then card["Charge"..i]=math.min(card["Charge"..i], getrechargetime("Charge")-1) end
  end
  gACTION = ""
  tCLICK = {}
  refreshmoves()
  if checkvictory() and AIsimulation==false then   -- Check victory condition
    gENDING = tSETTINGS.FPS * 4
    addpopup("P"..pl.." has won the game")
  end
end

function docharge(x, y)  -- perform a charge towards this space
  local a = tCLICK[#tCLICK].Player    -- Friendly
  local b = 3 - a                     -- Foe
  local card = tCOMBAT[a][tCLICK[#tCLICK].Index]
  local power = getpower(card)
  local ox, oy, oz = getcubecoord(card.GX, card.GY)
  local tx, ty, tz = getcubecoord(x, y)
  local steps = getcubedist(ox, oy, oz, tx, ty, tz)  -- How many spaces did they charge?
  local dx, dy, dz = (tx-ox)/5, (ty-oy)/5, (tz-oz)/5   -- One will be positive, one will be negative, one will be zero
  local i, j, cx, cy, cz, p, q
  if dx>=0 then dx = math.ceil(dx) else dx = math.floor(dx) end  -- Now +1, 0 or -1
  if dy>=0 then dy = math.ceil(dy) else dy = math.floor(dy) end  -- Now +1, 0 or -1
  if dz>=0 then dz = math.ceil(dz) else dz = math.floor(dz) end  -- Now +1, 0 or -1
  if steps > 0 and next(tCOMBAT[b])~=nil then
    for i = 1, steps, 1 do
      cx, cy, cz = ox+(i*dx), oy+(i*dy), oz+(i*dz)
      p, q = getflatcoord(cx, cy, cz)
      for j = #tCOMBAT[b], 1, -1 do
        if tCOMBAT[b][j].GX == p and tCOMBAT[b][j].GY == q then damage(tCOMBAT[b][j], "Combat", b, j, power) end
      end
    end
  end
  addpopup(card.Name.." charges")
  card.OriginalX, card.OriginalY = card.X, card.Y
  card.TargetX, card.TargetY = gridtoscreen(x, y)
  card.Animate = true
  card.AniCount = tSETTINGS.AniTime
  tHEX[card.GY][card.GX].Player=0
  card.GX, card.GY = x, y
  card.Steps = card.Steps + getreach(card)
  tHEX[card.GY][card.GX].Player=tCLICK[#tCLICK].Player
  gACTION = ""
  tCLICK = {}
  refreshmoves()
  if checkvictory() and AIsimulation == false then  -- Check victory condition
    gENDING = tSETTINGS.FPS * 4
    addpopup("P"..pl.." has won the game")
  end
end

function dotarget()  -- perform action on target (e.g. attack, play action)
  local a, b = tCLICK[#tCLICK].Player, tCLICK[1].Player
  local x, y, d, bx, by, bd    -- Used in spawn ant to find closest non-occupied point
  local origin, target
  local i
  if tCLICK[#tCLICK].Source == "Hand" then origin = tHAND[a][tCLICK[#tCLICK].Index]
  elseif tCLICK[#tCLICK].Source == "Combat" then origin = tCOMBAT[a][tCLICK[#tCLICK].Index]
  elseif tCLICK[#tCLICK].Source == "Grave" then origin = tGRAVE[a][tCLICK[#tCLICK].Index]
  elseif tCLICK[#tCLICK].Source == "Deck" then origin = tDECK[a][tCLICK[#tCLICK].Index]
  else end
  if tCLICK[1].Source == "Hand" then target = tHAND[b][tCLICK[1].Index]
  elseif tCLICK[1].Source == "Combat" then target = tCOMBAT[b][tCLICK[1].Index]
  elseif tCLICK[1].Source == "Grave" then target = tGRAVE[b][tCLICK[1].Index]
  elseif tCLICK[1].Source == "Deck" then target = tDECK[b][tCLICK[1].Index]
  else end
  if gACTION == "Move/Attack" then
    dobattle(origin, tCLICK[#tCLICK].Source, a, tCLICK[#tCLICK].Index, target, tCLICK[1].Source, b, tCLICK[1].Index)
    if gethasability(origin, "Energetic") and math.random(100)<=20 then addfloatytext(origin, "+Energetic+")
    else origin.Steps = origin.Steps + getreach(origin) end
  elseif gACTION == "Cage" then
    addpopup(target.Name.." is caged")
    target.Effect[1] = math.max(target.Effect[1], 4)
    target.Effect[2] = math.max(target.Effect[2], 4)
    addfloatytext(target, "-Caged-")
  elseif gACTION == "Calm Mammal" then
    addpopup(target.Name.." is calmed")
    target.Effect[4] = 0
    addfloatytext(target, "+Calmed+")
  elseif gACTION == "Charm Mammal" then
    addpopup(target.Name.." is charmed")
    target.Effect[4] = math.max(target.Effect[4], 2)
    addfloatytext(target, "-Charmed-")
  elseif gACTION == "Cure Mild Poison" then
    addpopup(target.Name.." takes a remedy")
    target.Effect[11] = 0
    if target.Effect[12]>0 then 
      target.Effect[11]=target.Effect[12]
      target.Effect[12]=0
    end
    if target.Effect[13]>0 then
      target.Effect[12]=target.Effect[13]
      target.Effect[13]=0
    end
    addfloatytext(target, "+Cure Poison+")
  elseif gACTION == "Kitten Scratch" then
    addpopup(target.Name.." is scratched")
    addfloatytext(target, "-Scratched-")
    damage(target, "Combat", b, tCLICK[1].Index, 1)
  elseif gACTION == "Mesmerise" then
    if math.random(100)<=80 then
      addpopup(target.Name.." is mesmerised")
      target.Effect[1] = math.max(target.Effect[1], 2)
      addfloatytext(target, "-Distracted-")
    end
  elseif gACTION == "Paralysis (Minor)" then
    addpopup(target.Name.." is paralysed by "..origin.Name)
    target.Effect[3] = math.max(target.Effect[3], 4)
    addfloatytext(target, "-Paralysis-")
  elseif gACTION == "Poison (Minor)" then
    addpopup(target.Name.." is poisoned")
    target.Effect[11] = 100
    addfloatytext(target, "-Poisoned-")
  elseif gACTION == "Snake Coil" then
    if math.random(100)<=75 then
      addpopup(origin.Name.." wraps around "..target.Name)
      target.Effect[3] = math.max(target.Effect[3], 4)
      addfloatytext(target, "-Limp-")
    else addfloatytext(target, "+Escaped+") end
  elseif gACTION == "Snatch Card" then
    if math.random(100)<=75 then
      addpopup("P"..a.." snatches a card from P"..b)
      addfloatytext(target, "+Snatched+")
      target.Animate = true
      target.AniCount = tSETTINGS.AniTime
      target.OriginalX, target.OriginalY = target.X, target.Y
      target.TargetX, target.TargetY = xHAND[a], yHAND[a]
      target.Steps = 0
      if tCONTROL[a]=="Player" then target.FaceUp = true
      else target.FaceUp = false end
      target.Clickable = true
      table.insert(tHAND[a], deepcopy(target))
      table.remove(tHAND[b], tCLICK[1].Index)
      arrangehand(a)
      arrangehand(b)
    else addfloatytext(target, "-Missed-") end
  elseif gACTION == "Spawn Ant" then
    addpopup("P"..a.." spawns another ant")
    bx, by, bd = 0, 0, math.huge
    for y = 1, gGRID_HEIGHT, 1 do
      for x = 1, gGRID_WIDTH, 1 do
        if tHEX[y][x].Player==0 then
          d = getflatdist(x, y, target.GX, target.GY)
          if d<bd then bx, by, bd = x, y, d end
        end
      end
    end
    if bx~=0 and by~=0 then
      y = tCLICK[1].Player
      tHEX[by][bx].Player = y
      table.insert(tCOMBAT[y], deepcopy(target))
      x = #tCOMBAT[y]
      tCOMBAT[y][x].GX, tCOMBAT[y][x].GY = bx, by
      tCOMBAT[y][x].X, tCOMBAT[y][x].Y = gridtoscreen(bx, by)
    end
  elseif gACTION == "Sprint" then
    target.Effect[15] = math.max(target.Effect[15], 2)
    addpopup(target.Name.." sprints")
  elseif gACTION == "Thorns" then
    addpopup("P"..a.." places some thorns")
    if next(tCOMBAT[gTURN])~=nil then
      for i = #tCOMBAT[gTURN], 1, -1 do
        if getisadjacent(tCOMBAT[gTURN][i], target) and tCOMBAT[gTURN][i].Size~="Tiny" then 
          addfloatytext(tCOMBAT[gTURN][i], "-Thorns-")
          if tCOMBAT[gTURN][i].Effect[2]>0 then
            addfloatytext(tCOMBAT[gTURN][i], "-Revealed-")
            tCOMBAT[gTURN][i].Effect[2]=0
          end
          damage(tCOMBAT[gTURN][i], "Combat", gTURN, i, 2)
        end
      end
    end
    if next(tCOMBAT[3-gTURN])~=nil then
      for i = #tCOMBAT[3-gTURN], 1, -1 do
        if getisadjacent(tCOMBAT[3-gTURN][i], target) and tCOMBAT[3-gTURN][i].Size~="Tiny" then
          addfloatytext(tCOMBAT[3-gTURN][i], "-Thorns-")
          if tCOMBAT[3-gTURN][i].Effect[2]>0 then
            addfloatytext(tCOMBAT[3-gTURN][i], "-Revealed-")
            tCOMBAT[3-gTURN][i].Effect[2]=0
          end
          damage(tCOMBAT[3-gTURN][i], "Combat", 3-gTURN, i, 2)
        end
      end
    end
    if target.Size~="Tiny" then damage(target, "Combat", tCLICK[1].Player, tCLICK[1].Index, 2) end
  elseif gACTION == "Web (Small)" then
    addpopup(target.Name.." is trapped in a web")
    addfloatytext(target, "-Webbed-")
    target.Effect[3] = math.max(target.Effect[3], 2)
  else end
end

function dobattle(att_card, att_source, att_player, att_index, def_card, def_source, def_player, def_index)  -- battle between two cards
  local att_power, def_power = getpower(att_card), getpower(def_card)
  local att_dead, def_dead
  local gx, gy = def_card.GX, def_card.GY
  addpopup(att_card.Name.." attacks "..def_card.Name)
  att_card.Effect[2] = 0
  if att_power>0 then
    if gethasability(def_card, "Evasive") and math.random(100)<=25 then addfloatytext(def_card, "+Evaded+")
    elseif gethasability(def_card, "Trickery") and math.random(100)<=50 then addfloatytext(def_card, "+Trickery+")
    elseif gethasability(def_card, "Imitate Call") and (att_card.Class=="Mammal" or att_card.Class=="Bird") and math.random(100)<=50 then addfloatytext(def_card, "+Imitate Call+")
    else def_dead = damage(def_card, def_source, def_player, def_index, att_power) end
  end
  if def_power>0 and def_card.Effect[3]==0 then
    if gethasability(att_card, "Evasive") and math.random(100)<=25 then addfloatytext(att_card, "+Evaded+")
    else att_dead = damage(att_card, att_source, att_player, att_index, def_power) end
  end
  if att_dead == false and gethasability(att_card, "Single Use") then
    addfloatytext(att_card, "-Single Use-")
    att_dead = damage(att_card, att_source, att_player, att_index, getkilldmg(att_card))
  end
  if def_dead == false and gethasability(def_card, "Single Use") then
    addfloatytext(def_card, "-Single Use-")
    def_dead = damage(def_card, def_source, def_player, def_index, getkilldmg(def_card))
  end
  if att_dead == false and def_dead == true and gethasability(att_card, "Territorial") then
    att_card.Animate = true
    att_card.AniTime = tSETTINGS.AniTime
    att_card.OriginalX, att_card.OriginalY = att_card.X, att_card.Y
    att_card.TargetX, att_card.TargetY = gridtoscreen(gx, gy)
    tHEX[att_card.GY][att_card.GX].Player = 0
    tHEX[gy][gx] = att_player
    att_card.GX, att_card.GY = gx, gy
    if checkvictory() and AIsimulation == false then
      gENDING = tSETTINGS.FPS * 4
      addpopup("P"..att_player.." has won the game")
    end
  end
end

function damage(card, loc, pl, index, amount)
  local i
  amount = math.max(amount - getarmour(card), 0)
  if amount > 0 then addfloatytext(card, "-"..amount)
  else addfloatytext(card, "0") end
  card.Wounds = card.Wounds + amount
  if AIsimulation==false then writelog("P"..pl.." "..card.Name.." suffers "..amount.." damage") end
  if gethealth(card)<=0 then
    if AIsimulation==false then writelog("P"..pl.." "..card.Name.." dies") end
    if gethasability(card, "Regrow") and math.random(100)<=50 then
      card.Wounds = 0
      addfloatytext(card, "+Regrows+")
      card.Effect[3] = math.max(card.Effect[3], 2)
    else 
      sendcard("Grave", card, loc, pl, index)
      if next(tCOMBAT[pl])~=nil then
        for i = 1, #tCOMBAT[pl], 1 do
          if gethasability(tCOMBAT[pl][i], "Carrion") and tCOMBAT[pl][i].Wounds>0 then
            addfloatytext(tCOMBAT[pl][i], "+Carrion+")
            tCOMBAT[pl][i].Wounds = math.max(tCOMBAT[pl][i].Wounds - math.ceil(tCOMBAT[pl][i].MaxHealth/2), 0)
          end
        end
      end
    end
    return true
  else
    return false
  end
end

function refreshmoves()  -- refresh the grid with how many moves it is away
  local dx, dy, dz = {0, 1, 1, 0, -1, -1}, {1, 0, -1, -1, 0, 1}, {-1, -1, 0, 1, 1, 0}      -- All six possible directions
  local cx, cy, cz, cm, ox, oy, oz, om
  local total_move, card, pl
  local i, j, x, y
  for y = 1, gGRID_HEIGHT, 1 do                                                               -- Set everywhere as a lot of moves to get there!
    for x = 1, gGRID_WIDTH, 1 do
      tHEX[y][x].Moves = math.huge
    end
  end
  if next(tCLICK)~=nil then
    card, pl = tCLICK[#tCLICK].Card, tCLICK[#tCLICK].Player                               -- Origin card was the first one clicked
    cx, cy, cz = getcubecoord(card.GX, card.GY)                                           -- Where do I start?
    if gACTION == "Move/Attack" then                                                      -- Pathfinding for move/attack
      total_move = getreach(card)                                                          -- How many move points can I spend in total?
      ox, oy, oz, om = {cx}, {cy}, {cz}, {0}                                                -- Insert those details as the first option
      while next(ox)~=nil or next(oy)~=nil or next(oz)~=nil do
        cx, cy, cz, cm = ox[1], oy[1], oz[1], om[1]                                          -- Current cube coordinates and points spent so far
        x, y = getflatcoord(cx, cy, cz)                                                     -- Converted into flat coords
        table.remove(ox, 1)
        table.remove(oy, 1)
        table.remove(oz, 1)
        table.remove(om, 1)
        for i = 1, 6, 1 do
          x, y = getflatcoord(cx+dx[i], cy+dy[i], cz+dz[i])                                  -- Flat coords to try (six directions)
          if x>=1 and x<=gGRID_WIDTH and y>=1 and y<=gGRID_HEIGHT and tHEX[y][x].Player~=-1 and tHEX[y][x].Player~=pl and cm+1<tHEX[y][x].Moves and cm+1<=total_move then   -- New best
            tHEX[y][x].Moves = cm+1
            table.insert(ox, cx+dx[i])
            table.insert(oy, cy+dy[i])
            table.insert(oz, cz+dz[i])
            table.insert(om, cm+1)
          end
        end
      end
      cx, cy, cz = getcubecoord(card.GX, card.GY)
      if gethasability(card, "Jump") then    -- Jump over one space
        for i = 1, 6, 1 do
          x, y = getflatcoord(cx+(2*dx[i]), cy+(2*dy[i]), cz+(2*dz[i]))
          if x>=1 and x<=gGRID_WIDTH and y>=1 and y<=gGRID_HEIGHT and tHEX[y][x].Player==0 then tHEX[y][x].Moves = 1 end
        end
      end
      if gethasability(card, "Scuttle") then   -- Move any spaces sideways to a free space
        for x = 1, gGRID_WIDTH, 1 do
          if tHEX[card.GY][x].Player == 0 then tHEX[card.GY][x].Moves = 1 end
        end
      end
    elseif gACTION == "Charge" then   -- Charge in a straight line to a free space, any number of spaces
      for i = 1, gGRID_WIDTH, 1 do
        for j = 1, 6, 1 do
          x, y = getflatcoord(cx+(i*dx[j]), cy+(i*dy[j]), cz+(i*dz[j]))
          if x>=1 and x<=gGRID_WIDTH and y>=1 and y<=gGRID_HEIGHT and tHEX[y][x].Player==0 then tHEX[y][x].Moves = 1 end
        end
      end
    else end
  end
end

function love.update()  -- Runs every frame
  updategeneral()
  if gGAMEMODE == "game" then updategame() end
  musicupdate()
  collectgarbage()
  if love.timer.getTime()-gTIMER<=(1/tSETTINGS.FPS) then love.timer.sleep((1/tSETTINGS.FPS)-(love.timer.getTime()-gTIMER)) end  -- frame rate
  gTIMER = love.timer.getTime()  -- new timer
end

function updategeneral()  -- Things that are updated every frame in all gamemodes
  if love.mouse.isDown(1) then gMOUSEGRAB = gMOUSEGRAB + 1 end
  if gFADE > 0 then  -- fade in graphics
    gFADE = gFADE - gFADESPEED
    if gFADE < 0 then gFADE = 0 end
  end
  gFLASH = gFLASH + gFLASH_z  -- flash counter
  if gFLASH > 255 then  -- this affects everything that flashes (same speed)
    gFLASH = gFLASH_max
    gFLASH_z = 0 - gFLASH_z
  elseif gFLASH < gFLASH_min then
    gFLASH = gFLASH_min
    gFLASH_z = 0 - gFLASH_z
  else end
  if gPOPUP_COUNT > 0 then  -- Pop up message counter
    gPOPUP_COUNT = gPOPUP_COUNT - 1
    if gPOPUP_COUNT <= 0 then gPOPUP_MSG = "" end
  end
end

function updategame()  -- Things that are updated every frame only whilst on the main game screen
  local i, j
  for i = 1, 2, 1 do   
    if next(tDECK[i]) ~= nil then  -- Card animations (deck)
      for j = 1, #tDECK[i], 1 do
        if tDECK[i][j].Animate == true then animatecard(tDECK[i][j]) end
        if next(tDECK[i][j].FloatyText) ~= nil then animatetext(tDECK[i][j]) end
      end
    end
    if next(tHAND[i]) ~= nil then  -- Card animations (hand)
      for j = 1, #tHAND[i], 1 do
        if tHAND[i][j].Animate == true then animatecard(tHAND[i][j]) end
        if next(tHAND[i][j].FloatyText) ~= nil then animatetext(tHAND[i][j]) end
      end
    end
    if next(tGRAVE[i]) ~= nil then  -- Card animations (graveyard)
      for j = 1, #tGRAVE[i], 1 do
        if tGRAVE[i][j].Animate == true then animatecard(tGRAVE[i][j]) end
        if next(tGRAVE[i][j].FloatyText) ~= nil then animatetext(tGRAVE[i][j]) end
      end
    end
    if next(tCOMBAT[i]) ~= nil then  -- Card animations (combat)
      for j = 1, #tCOMBAT[i], 1 do
        if tCOMBAT[i][j].Animate == true then animatecard(tCOMBAT[i][j]) end
        if next(tCOMBAT[i][j].FloatyText) ~= nil then animatetext(tCOMBAT[i][j]) end
      end
    end
  end
  if gENDING > 0 then 
    gENDING  = gENDING - 1
    if gENDING == 1 then
      gFADE = 255
      endgame()
    end
  end
  if tCONTROL[gTURN] == "AI" and gENDING == 0 then AI() end -- Call the A.I.
end

function animatecard(card)
  card.X = card.X + ((card.TargetX - card.OriginalX) / tSETTINGS.AniTime)
  card.Y = card.Y + ((card.TargetY - card.OriginalY) / tSETTINGS.AniTime)
  card.AniCount = card.AniCount - 1
  if card.AniCount <= 0 then
    playsound(3)
    card.AniCount = 0
    card.X = card.TargetX
    card.Y = card.TargetY
    card.Animate = false
  end
end

function animatetext(card)
  local i = 0
  repeat
    i = i + 1
    card.FloatyText[i].Timer = card.FloatyText[i].Timer + 1
  until i==#card.FloatyText or card.FloatyText[i].Timer <= 40
  if card.FloatyText[1].Timer >= tSETTINGS.FloatTime then table.remove(card.FloatyText, 1) end
end

function love.draw()  -- Draw everything in, based on gamemode
  local mx, my = love.mouse.getX()/gRATIOX, love.mouse.getY()/gRATIOY
  love.graphics.scale(gRATIOX, gRATIOY)
  drawbackground()
  if _G["draw"..gGAMEMODE]~=nil then _G["draw"..gGAMEMODE]() end
  drawpopuptext()
  if gFADE > 0 then 
    love.graphics.setColor(0, 0, 0, gFADE)
    love.graphics.rectangle("fill", 0, 0, 800, 600)
  else end
end

function drawbackground()
  love.graphics.setColor(255, 255, 255, 255)
  love.graphics.draw(iBACKGROUND, 0, 0)
end

function drawpopuptext()
  local w, wrappedtext, n, h
  if gPOPUP_COUNT > 0 then
    love.graphics.setFont(fHEADING)
    h = fHEADING:getHeight()
    w, wrappedtext = fHEADING:getWrap(gPOPUP_MSG, 390)
    if next(wrappedtext)~=nil then n=#wrappedtext else n = 0 end
    love.graphics.setColor(0, 0, 0, 255)
    love.graphics.rectangle("fill", 400-(w/2)-5, 300-(n*h/2)-5, w+10, (n*h)+10)
    love.graphics.setColor(255, 255, 255, 255)
    love.graphics.rectangle("line", 400-(w/2)-5, 300-(n*h/2)-5, w+10, (n*h)+10)
    love.graphics.printf(gPOPUP_MSG, 400-(w/2), 300-(n*h/2), w, "center")
  end
end

function drawanimatedtext()
  local i, j, k
  local x, y, z, t
  for j = 1, 2, 1 do
    if next(tDECK[j])~=nil then
      for i = 1, #tDECK[j], 1 do
        if next(tDECK[j][i].FloatyText)~=nil then
          for k = 1, #tDECK[j][i].FloatyText, 1 do
            x = tDECK[j][i].FloatyText[k].X
            y = tDECK[j][i].FloatyText[k].Y-tDECK[j][i].FloatyText[k].Timer
            z = math.floor(255*((tSETTINGS.FloatTime-tDECK[j][i].FloatyText[k].Timer)/tSETTINGS.FloatTime))
            t = tDECK[j][i].FloatyText[k].Text
            if string.find(t, "+") then drawshadowtext(t, x-100, y, 200, "center", fSUBHEADING, 0, 255, 0, z)
            elseif string.find(t, "-") then drawshadowtext(t, x-100, y, 200, "center", fSUBHEADING, 255, 0, 0, z)
            else drawshadowtext(t, x-100, y, 200, "center", fSUBHEADING, 255, 255, 255, z) end
          end
        end
      end
    end
    if next(tHAND[j])~=nil then
      for i = 1, #tHAND[j], 1 do
        if next(tHAND[j][i].FloatyText)~=nil then
          for k = 1, #tHAND[j][i].FloatyText, 1 do
            x = tHAND[j][i].FloatyText[k].X
            y = tHAND[j][i].FloatyText[k].Y-tHAND[j][i].FloatyText[k].Timer
            z = math.floor(255*((tSETTINGS.FloatTime-tHAND[j][i].FloatyText[k].Timer)/tSETTINGS.FloatTime))
            t = tHAND[j][i].FloatyText[k].Text
            if string.find(t, "+") then drawshadowtext(t, x-100, y, 200, "center", fSUBHEADING, 0, 255, 0, z)
            elseif string.find(t, "-") then drawshadowtext(t, x-100, y, 200, "center", fSUBHEADING, 255, 0, 0, z)
            else drawshadowtext(t, x-100, y, 200, "center", fSUBHEADING, 255, 255, 255, z) end
          end
        end
      end
    end
    if next(tCOMBAT[j])~=nil then
      for i = 1, #tCOMBAT[j], 1 do
        if next(tCOMBAT[j][i].FloatyText)~=nil then
          for k = 1, #tCOMBAT[j][i].FloatyText, 1 do
            x = tCOMBAT[j][i].FloatyText[k].X
            y = tCOMBAT[j][i].FloatyText[k].Y-tCOMBAT[j][i].FloatyText[k].Timer
            z = math.floor(255*((tSETTINGS.FloatTime-tCOMBAT[j][i].FloatyText[k].Timer)/tSETTINGS.FloatTime))
            t = tCOMBAT[j][i].FloatyText[k].Text
            if string.find(t, "+") then drawshadowtext(t, x-100, y, 200, "center", fSUBHEADING, 0, 255, 0, z)
            elseif string.find(t, "-") then drawshadowtext(t, x-100, y, 200, "center", fSUBHEADING, 255, 0, 0, z)
            else drawshadowtext(t, x-100, y, 200, "center", fSUBHEADING, 255, 255, 255, z) end
          end
        end
      end
    end
    if next(tGRAVE[j])~=nil then
      for i = 1, #tGRAVE[j], 1 do
        if next(tGRAVE[j][i].FloatyText)~=nil then
          for k = 1, #tGRAVE[j][i].FloatyText, 1 do
            x = tGRAVE[j][i].FloatyText[k].X
            y = tGRAVE[j][i].FloatyText[k].Y-tGRAVE[j][i].FloatyText[k].Timer
            z = math.floor(255*((tSETTINGS.FloatTime-tGRAVE[j][i].FloatyText[k].Timer)/tSETTINGS.FloatTime))
            t = tGRAVE[j][i].FloatyText[k].Text
            if string.find(t, "+") then drawshadowtext(t, x-100, y, 200, "center", fSUBHEADING, 0, 255, 0, z)
            elseif string.find(t, "-") then drawshadowtext(t, x-100, y, 200, "center", fSUBHEADING, 255, 0, 0, z)
            else drawshadowtext(t, x-100, y, 200, "center", fSUBHEADING, 255, 255, 255, z) end
          end
        end
      end
    end
  end  
end

function drawtitle()  -- Draw the title menu
  drawshadowtext("Hexanimals", 275, 275, 250, "center", fTITLE, 255, 215, 0)
end

function drawmenu()  -- Draw the main menu
  drawshadowtext("Menu", 330, 75, 140, "center", fTITLE, 255, 215, 0)
  drawshadowtext("Campaign", 300, 200, 200, "center", fHEADING, 255, 255, 255, 255, "Yes", "No", "No", gFLASH)
  drawshadowtext("Pass and Play", 300, 250, 200, "center", fHEADING, 255, 255, 255, 255, "Yes", "No", "No", gFLASH)
  drawshadowtext("Deck", 300, 300, 200, "center", fHEADING, 255, 255, 255, 255, "Yes", "No", "No", gFLASH)
  drawshadowtext("Settings", 300, 350, 200, "center", fHEADING, 255, 255, 255, 255, "Yes", "No", "No", gFLASH)
  drawshadowtext("Quit", 300, 400, 200, "center", fHEADING, 255, 255, 255, 255, "Yes", "No", "No", gFLASH)
  drawshadowtext("Music: bensound.com", 530, 550, 250, "right", fSUBHEADING)
end

function drawsettings()  -- Draw the settings page
  local list = {"FPS", "Sound", "Music", "MusicVol", "Gridsize", "P1COLOUR", "P2COLOUR"}
  local i
  drawshadowtext("Settings", 300, 75, 200, "center", fTITLE, 255, 215, 0)
  for i = 1, #list, 1 do
    if i<=8 then 
      drawshadowtext(list[i], 0, 100+(i*40), 190, "right", fHEADING)
      if tSETTINGS[list[i]]~=nil and (list[i]=="P1COLOUR" or list[i]=="P2COLOUR") then
        drawshadowtext(tSETTINGS[list[i]], 210, 100+(i*40), 190, "left", fHEADING, tCOLOUR[tSETTINGS[list[i]]].R, tCOLOUR[tSETTINGS[list[i]]].G, tCOLOUR[tSETTINGS[list[i]]].B, 255, "Yes", "No", "No", gFLASH)
      elseif tSETTINGS[list[i]]~=nil then 
        drawshadowtext(tSETTINGS[list[i]], 210, 100+(i*40), 190, "left", fHEADING, 255, 215, 0, 255, "Yes", "No", "No", gFLASH)
      else end
    else
      drawshadowtext(list[i], 400, 100+((i-8)*40), 190, "right", fHEADING)
      if tSETTINGS[list[i]]~=nil and (list[i]=="P1COLOUR" or list[i]=="P2COLOUR") then
        drawshadowtext(tSETTINGS[list[i]], 210, 100+((i-8)*40), 190, "left", fHEADING, tCOLOUR[tSETTINGS[list[i]]].R, tCOLOUR[tSETTINGS[list[i]]].G, tCOLOUR[tSETTINGS[list[i]]].B, 255, "Yes", "No", "No", gFLASH)
      elseif tSETTINGS[list[i]]~=nil then 
        drawshadowtext(tSETTINGS[list[i]], 610, 100+((i-8)*40), 190, "left", fHEADING, 255, 215, 0, 255, "Yes", "No", "No", gFLASH)
      else end
    end
  end
  drawshadowtext("Back/Save", 300, 475, 200, "center", fHEADING, 255, 255, 255, 255, "Yes", "No", "No", gFLASH)
end

function drawmap()  -- Draw the campaign map
  local i
  love.graphics.setColor(255, 255, 255, 255)
  love.graphics.draw(iMAP[tPLAYER.LEVEL], 0, 0)
  drawshadowtext("Back", 720, 560, 70, "center", fHEADING, 255, 255, 255, 255, "Yes", "Yes", "Yes", gFLASH)
  for i = 1, tPLAYER.LEVEL, 1 do
    love.graphics.setColor(255, 255, 255, 255)
    love.graphics.circle("fill", tMAP[i].X, tMAP[i].Y, 6)
    if i == gSELECT then love.graphics.circle("line", tMAP[i].X, tMAP[i].Y, 6+(gFLASH/50)) end
    if i == gSUBSELECT then 
      drawshadowtext("Easy", tMAP[i].X+20, tMAP[i].Y-36, 50, "center", fSMALL, 0, 255, 0, 255, "Yes", "Yes", "Yes", gFLASH)
      drawshadowtext("Medium", tMAP[i].X+20, tMAP[i].Y-12, 50, "center", fSMALL, 255, 140, 0, 255, "Yes", "Yes", "Yes", gFLASH)
      drawshadowtext("Hard", tMAP[i].X+20, tMAP[i].Y+12, 50, "center", fSMALL, 255, 0, 0, 255, "Yes", "Yes", "Yes", gFLASH)
    end
  end
end

function drawnewturn()
  drawshadowtext("P"..gTURN..": Your Turn", 200, 275, 400, "center", fTITLE, 255, 215, 0, 255)
end

function drawgame()
  local i
  drawbackground()
  love.graphics.setColor(255, 255, 255, 255)
  for i = 1, 2, 1 do
    love.graphics.draw(iICON[17], xGRAVE[i], yGRAVE[i], 0, 1, 1, iICON[17]:getWidth()/2, iICON[17]:getHeight()/2)
    checkhover(i)    
  end
  drawgrid()
  for i = 1, 2, 1 do
    drawdeck(i)
    drawgrave(i)
    drawhand(i)
    drawcombat(i)
  end
  for i = 1, 2, 1 do
    drawhover(i)
  end
  drawanimatedtext()
  drawgui()
  drawselectedcard()
  drawheldcard()
end

function drawcard(card, x, y, width, height, angle, alpha)
  angle = angle or 0
  alpha = alpha or 255
  local h = gethealth(card)
  local c = card.Tier
  local a = getpower(card)
  local r, g, b = card.R or 255, card.G or 255, card.B or 255
  local sx, sy = width/300, height/420
  local name = card.Name
  local i, j, recharge
  local text = card.Class
  if card.Type=="Hero" then name = name.." (Hero)" end
  if card.Type=="Ally" or card.Type=="Hero" then text=text..", "..card.Size end
  angle = math.rad(angle)
  if card.FaceUp == nil or card.FaceUp == true then
    love.graphics.setColor(r, g, b, alpha)
    love.graphics.draw(iFRONT, x, y, angle, sx, sy, 150, 210)
    love.graphics.setColor(255, 255, 255, alpha)
    love.graphics.draw(iIMAGE[card.Pic], x, y, angle, sx, sy, 80, 165)
    love.graphics.draw(iSQUARE, x, y, angle, sx, sy, 80, 165)
    love.graphics.draw(iICON[10], x, y, angle, sx, sy, -78, -132)
    love.graphics.draw(iICON[9], x, y, angle, sx, sy, 144, 204)
    love.graphics.setFont(fTITLE)
    love.graphics.setColor(0, 0, 0, alpha)
    if a > 0 then love.graphics.printf(a, x, y, 68, "center", angle, sx, sy, -78, -156)
    else love.graphics.printf("-", x, y, 68, "center", angle, sx, sy, -78,-156) end
    if c > 0 then love.graphics.printf(c, x, y, 60, "center", angle, sx, sy, 144, 190)
    else love.graphics.printf("-", x, y, 60, "center", angle, sx, sy, 144, 190) end
    love.graphics.setColor(255, 255, 255, alpha)
    if a > 0 then love.graphics.printf(a, x, y, 68, "center", angle, sx, sy, -80, -158)
    else love.graphics.printf("-", x, y, 68, "center", angle, sx, sy, -80, -158) end
    if c > 0 then love.graphics.printf(c, x, y, 60, "center", angle, sx, sy, 146, 192)
    else love.graphics.printf("-", x, y, 60, "center", angle, sx, sy, 146, 192) end
    love.graphics.setFont(fHEADING)
    love.graphics.setColor(0, 0, 0, alpha)
    love.graphics.printf(name, x, y, 230, "left", angle, sx, sy, 76, 200)
    love.graphics.printf(text, x, y, 160, "center", angle, sx, sy, 78, -144)
    love.graphics.setColor(255, 255, 255, alpha)
    love.graphics.printf(name, x, y, 230, "left", angle, sx, sy, 78, 202)
    love.graphics.printf(text, x, y, 160, "center", angle, sx, sy, 80, -142)
    if card.MaxHealth > 0 then
      if h > card.MaxHealth*0.7 then love.graphics.setColor(0, 255, 0, alpha) 
      elseif h > card.MaxHealth*0.4 then love.graphics.setColor(255, 150, 0, alpha)
      else love.graphics.setColor(255, 0, 0, alpha) end
      love.graphics.draw(iICON[11], x, y, angle, sx, sy, 146, -146) 
      love.graphics.setColor(0, 0, 0, alpha)
      love.graphics.printf(h, x, y, 40, "center", angle, sx, sy, 149, -148)
      love.graphics.printf("/", x, y, 60, "center", angle, sx, sy, 143, -156)
      love.graphics.printf(card.MaxHealth, x, y, 40, "center", angle, sx, sy, 119, -176)
      love.graphics.setColor(255, 255, 255, alpha)
      love.graphics.printf(h, x, y, 40, "center", angle, sx, sy, 151, -146)
      love.graphics.printf("/", x, y, 60, "center", angle, sx, sy, 144, -155)
      love.graphics.printf(card.MaxHealth, x, y, 40, "center", angle, sx, sy, 121, -174)
    end
    love.graphics.setFont(fDEFAULT)
    love.graphics.setColor(0, 0, 0, alpha)
    love.graphics.printf(card.Tag, x, y, 270, "left", angle, sx, sy, 134, -8)
    for i = 1, 4, 1 do
      if card["Spec"..i]~=nil then 
        recharge = getrechargetime(card["Spec"..i])
        if recharge == 0 then 
          love.graphics.printf(card["Spec"..i].." (automatic)", x, y, 270, "left", angle, sx, sy, 134, -41-(i*20))
        else 
          love.graphics.printf(card["Spec"..i].." (every "..recharge.." turns)", x, y, 270, "left", angle, sx, sy, 134, -41-(i*20))
        end
      end
    end
    love.graphics.setColor(255, 255, 255, alpha)
    love.graphics.printf(card.Tag, x, y, 270, "left", angle, sx, sy, 135, -7)
    love.graphics.setColor(255, 215, 0, alpha)
    for i = 1, 4, 1 do
      if card["Spec"..i]~=nil then 
        recharge = getrechargetime(card["Spec"..i])
        if recharge == 0 then
          love.graphics.printf(card["Spec"..i].." (automatic)", x, y, 270, "left", angle, sx, sy, 135, -40-(i*20))
        else
          love.graphics.printf(card["Spec"..i].." (every "..recharge.." turns)", x, y, 270, "left", angle, sx, sy, 135, -40-(i*20))
        end
      end
    end
    love.graphics.setColor(255, 255, 255, alpha)
    j = 0
    for i = 1, #tEFFECTS, 1 do
      if card.Effect~=nil and card.Effect[i]>0 then
        j = j + 1
        if j<=3 then
          love.graphics.draw(iICON[tEFFECTS[i].Pic], x, y, angle, sx, sy, 133, 185-(j*45))
        elseif j>3 and j<=6 then
          love.graphics.draw(iICON[tEFFECTS[i].Pic], x, y, angle, sx, sy, -95, 185-((j-3)*45))
        else end
      end
    end
  else
    love.graphics.setColor(255, 255, 255, 255)
    love.graphics.draw(iBACK, x, y, angle, sx, sy, 150, 210)
  end
end

function drawhand(pl)
  local i
  if next(tHAND[pl]) ~= nil then
    for i = 1, #tHAND[pl], 1 do
      if tHAND[pl][i].Hover == false then
        drawcard(tHAND[pl][i], tHAND[pl][i].X, tHAND[pl][i].Y, tSETTINGS.CardWidth, tSETTINGS.CardHeight, tHAND[pl][i].Angle)
      end
    end
  end
end

function checkhover(pl)
  local x, y = love.mouse.getX() / gRATIOX, love.mouse.getY() / gRATIOY
  local i, j, m, width
  local a, b
  local hover = 0
  if next(tDECK[pl]) ~= nil then
    for i = 1, #tDECK[pl], 1 do
      tDECK[pl][i].Hover = false
    end
    if hover == 0 and (gZOOMED == false or gDRAGGING == true) then
      if x>xDECK[pl]-(tSETTINGS.CardWidth/2)-12 and x<xDECK[pl]+(tSETTINGS.CardWidth/2)-12 and y>yDECK[pl]-(tSETTINGS.CardHeight/2)-12 and y<yDECK[pl]+(tSETTINGS.CardHeight/2)-12 and tDECK[pl][1].Clickable == true then
        tDECK[pl][1].Hover = true
        hover = 1
      end
    end
  end
  if next(tGRAVE[pl]) ~= nil then
    for i = 1, #tGRAVE[pl], 1 do
      tGRAVE[pl][i].Hover = false
    end
    if hover == 0 and (gZOOMED == false or gDRAGGING == true) then
      i = #tGRAVE[pl]
      if x>xGRAVE[pl]-(tSETTINGS.CardWidth/2) and x<xGRAVE[pl]+(tSETTINGS.CardWidth/2) and y>yGRAVE[pl]-(tSETTINGS.CardHeight/2) and y<yGRAVE[pl]+(tSETTINGS.CardHeight/2) and tGRAVE[pl][i].Clickable == true then
        tGRAVE[pl][i].Hover = true
        hover = i
      end
    end
  end
  if next(tHAND[pl]) ~= nil then
    for i = 1, #tHAND[pl], 1 do
      tHAND[pl][i].Hover = false
    end
    if hover == 0 and (gZOOMED == false or gDRAGGING == true) then
      m = (#tHAND[pl]+1)/2
      width = tSETTINGS.HandWidth-tSETTINGS.CardWidth
      for i = 1, #tHAND[pl], 1 do
        j = (i-m)/m
        if x>xHAND[pl]+(j*width)-(tSETTINGS.CardWidth/2) and x<xHAND[pl]+(j*width)+(tSETTINGS.CardWidth/2) and y>yHAND[pl]-(tSETTINGS.HandHeight/2) and y<yHAND[pl]+(tSETTINGS.HandHeight/2) and tHAND[pl][i].Clickable == true then
          hover = i
        end
      end
      if hover ~= 0 then tHAND[pl][hover].Hover=true end
    end
  end
  if next(tCOMBAT[pl]) ~= nil then
    for i = 1, #tCOMBAT[pl], 1 do
      tCOMBAT[pl][i].Hover = false
    end
    if hover == 0 and (gZOOMED == false or gDRAGGING == true) then
      for m = 1, #tCOMBAT[pl], 1 do
        a, b = screentogrid(x, y)
        if a==tCOMBAT[pl][m].GX and b==tCOMBAT[pl][m].GY and tCOMBAT[pl][m].Clickable == true then hover = m end
      end
      if hover ~= 0 then tCOMBAT[pl][hover].Hover=true end
    end
  end
end

function drawdeck(pl)
  local i, stack
  if next(tDECK[pl]) ~= nil then
    stack = math.min(#tDECK[pl], 5)
    for i = stack, 1, -1 do
      if tDECK[pl][i].Hover == false then drawcard(tDECK[pl][i], xDECK[pl]-((5-i)*3), yDECK[pl]-((5-i)*3), tSETTINGS.CardWidth, tSETTINGS.CardHeight) end 
    end
  end
end

function drawgrid()
  local mx, my = love.mouse.getX()/gRATIOX, love.mouse.getY()/gRATIOY
  local i, j, x, y, card, reach
  if next(tCLICK)~=nil and (gACTION=="Move/Attack" or gACTION=="Charge") then 
    card = tCLICK[#tCLICK].Card
    reach = getreach(card)
  end
  for j = 1, gGRID_HEIGHT, 1 do
    for i = 1, gGRID_WIDTH, 1 do
      x, y = gridtoscreen(i, j)
      if tHEX[j][i].Player==0 then drawhex(x, y, 60, 70, 255, 255, 255, 255, "Yes", 255, 255, 255, 100)
      elseif tHEX[j][i].Player==1 then drawhex(x, y, 60, 70, tCOLOUR[tSETTINGS.P1COLOUR].R, tCOLOUR[tSETTINGS.P1COLOUR].G, tCOLOUR[tSETTINGS.P1COLOUR].B, 255, "Yes", tCOLOUR[tSETTINGS.P1COLOUR].R, tCOLOUR[tSETTINGS.P1COLOUR].G, tCOLOUR[tSETTINGS.P1COLOUR].B, 100)
      elseif tHEX[j][i].Player==2 then drawhex(x, y, 60, 70, tCOLOUR[tSETTINGS.P2COLOUR].R, tCOLOUR[tSETTINGS.P2COLOUR].G, tCOLOUR[tSETTINGS.P2COLOUR].B, 255, "Yes", tCOLOUR[tSETTINGS.P2COLOUR].R, tCOLOUR[tSETTINGS.P2COLOUR].G, tCOLOUR[tSETTINGS.P2COLOUR].B, 100)
      else end
      if card ~= nil then
        if tHEX[j][i].Moves <= reach then drawhex(x, y, 60, 70, 255, 255, 255, 255, "Yes", 150, 150, 150, gFLASH) end
      end
      if j==1 and tHEX[j][i].Player == 0 then 
        love.graphics.setColor(tCOLOUR[tSETTINGS.P2COLOUR].R, tCOLOUR[tSETTINGS.P2COLOUR].G, tCOLOUR[tSETTINGS.P2COLOUR].B, 255)
        love.graphics.draw(iICON[8], x, y, 0, tSETTINGS.HexWidth/60, tSETTINGS.HexHeight/70, 20, 20)
      elseif j==gGRID_HEIGHT and tHEX[j][i].Player == 0 then 
        love.graphics.setColor(tCOLOUR[tSETTINGS.P1COLOUR].R, tCOLOUR[tSETTINGS.P1COLOUR].G, tCOLOUR[tSETTINGS.P1COLOUR].B, 255)
        love.graphics.draw(iICON[8], x, y, 0, tSETTINGS.HexWidth/60, tSETTINGS.HexHeight/70, 20, 20)
      else end
    end
  end
  if gTURN==1 then
    love.graphics.setColor(tCOLOUR[tSETTINGS.P1COLOUR].R, tCOLOUR[tSETTINGS.P1COLOUR].G, tCOLOUR[tSETTINGS.P1COLOUR].B, 255)
    love.graphics.draw(iICON[13], 105, 558)
  else
    love.graphics.setColor(tCOLOUR[tSETTINGS.P2COLOUR].R, tCOLOUR[tSETTINGS.P2COLOUR].G, tCOLOUR[tSETTINGS.P2COLOUR].B, 255)
    love.graphics.draw(iICON[13], 105, 22)
  end
end

function drawcombat(pl)
  local i
  if next(tCOMBAT[pl])~=nil then
    for i = 1, #tCOMBAT[pl], 1 do
      if tCOMBAT[pl][i].Hover == false then drawminipic(tCOMBAT[pl][i], tCOMBAT[pl][i].X, tCOMBAT[pl][i].Y, 0.6) end
    end
  end
end

function drawminipic(card, x, y, scale)
  local w = math.min(tSETTINGS.HexWidth*scale, tSETTINGS.HexHeight*scale)
  if gGAMEMODE=="game" and card.Effect[2]>0 then love.graphics.setColor(255, 255, 255, 150)
  elseif gGAMEMODE~="game" or getreach(card)>0 then love.graphics.setColor(255, 255, 255, 255)
  else love.graphics.setColor(150, 150, 150, 255) end
  love.graphics.draw(iIMAGE[card.Pic], x, y, 0, w/160, w/160, 80, 80)
  love.graphics.draw(iSQUARE, x, y, 0, w/160, w/160, 80, 80)
end

function drawgrave(pl)
  local i
  if next(tGRAVE[pl]) ~= nil then
    for i = 1, #tGRAVE[pl], 1 do
      if tGRAVE[pl][i].Hover == false then drawcard(tGRAVE[pl][i], tGRAVE[pl][i].X, tGRAVE[pl][i].Y, tSETTINGS.CardWidth, tSETTINGS.CardHeight, tGRAVE[pl][i].Angle) end 
    end
  end
end

function drawhover(pl)
  local i, x, y, w
  if next(tDECK[pl]) ~= nil then
    if tDECK[pl][1].Hover == true then
      if pl==1 then
        drawcard(tDECK[pl][1], xDECK[pl]+(tSETTINGS.CardWidth*0.2)-12, yDECK[pl]-(tSETTINGS.CardHeight*0.2)-12, tSETTINGS.CardWidth*1.4, tSETTINGS.CardHeight*1.4)
      else
        drawcard(tDECK[pl][1], xDECK[pl]+(tSETTINGS.CardWidth*0.2)-12, yDECK[pl]+(tSETTINGS.CardHeight*0.2)-12, tSETTINGS.CardWidth*1.4, tSETTINGS.CardHeight*1.4)
      end
    end
  end
  if next(tGRAVE[pl]) ~= nil then
    i = #tGRAVE[pl]
    if tGRAVE[pl][i].Hover == true then
      if pl==1 then
        drawcard(tGRAVE[pl][i], xGRAVE[pl]-(tSETTINGS.CardWidth*0.2), yGRAVE[pl]+(tSETTINGS.CardHeight*0.2), tSETTINGS.CardWidth*1.4, tSETTINGS.CardHeight*1.4, tGRAVE[pl][i].Angle)
      else
        drawcard(tGRAVE[pl][i], xGRAVE[pl]-(tSETTINGS.CardWidth*0.2), yGRAVE[pl]-(tSETTINGS.CardHeight*0.2), tSETTINGS.CardWidth*1.4, tSETTINGS.CardHeight*1.4, tGRAVE[pl][i].Angle)
      end
    end
  end
  if next(tHAND[pl]) ~= nil then
    for i = 1, #tHAND[pl], 1 do
      if tHAND[pl][i].Hover == true then
        if pl == 1 then drawcard(tHAND[pl][i], tHAND[pl][i].X, tHAND[pl][i].Y-(tSETTINGS.CardHeight*0.2), tSETTINGS.CardWidth*1.4, tSETTINGS.CardHeight*1.4, tHAND[pl][i].Angle)
        else drawcard(tHAND[pl][i], tHAND[pl][i].X, tHAND[pl][i].Y+(tSETTINGS.CardHeight*0.2), tSETTINGS.CardWidth*1.4, tSETTINGS.CardHeight*1.4, tHAND[pl][i].Angle) end
      end
    end
  end
  love.graphics.setColor(255, 255, 255, 255)
  if next(tCOMBAT[pl]) ~= nil then
    for i = 1, #tCOMBAT[pl], 1 do
      if tCOMBAT[pl][i].Hover == true then drawminipic(tCOMBAT[pl][i], tCOMBAT[pl][i].X, tCOMBAT[pl][i].Y, 0.8) end
    end
  end
end

function drawgui()
  local mx, my = love.mouse.getX()/gRATIOX, love.mouse.getY()/gRATIOY
  love.graphics.setColor(255, 255, 255, 255)
  if gDAYNIGHT>=6 then 
    love.graphics.draw(iICON[1], 560, 270)
    drawshadowtext(11-gDAYNIGHT, 560, 280, 60, "center", fTITLE, 255, 255, 255, 255)
    if mx>560 and mx<620 and my>280 and my<320 and gZOOMED==false then
      if gDAYNIGHT==10 then drawshadowtext("1 turn of daylight remaining", mx+25, my, 150, "left", fDEFAULT, 255, 255, 255, 255, "Yes", "Yes", "Yes")
      else drawshadowtext(11-gDAYNIGHT.." turns of daylight remaining", mx+25, my, 150, "left", fDEFAULT, 255, 255, 255, 255, "Yes", "Yes", "Yes") end
    end
  else 
    love.graphics.draw(iICON[2], 560, 270)
    drawshadowtext(6-gDAYNIGHT, 560, 280, 60, "center", fTITLE, 50, 0, 200, 255)
    if mx>560 and mx<620 and my>280 and my<320 and gZOOMED==false  then
      if gDAYNIGHT==5 then drawshadowtext("1 turn of night remaining", mx+25, my, 150, "left", fDEFAULT, 255, 255, 255, 255, "Yes", "Yes", "Yes")
      else drawshadowtext(6-gDAYNIGHT.." turns of night remaining", mx+25, my, 150, "left", fDEFAULT, 255, 255, 255, 255, "Yes", "Yes", "Yes") end
    end
  end
  love.graphics.setColor(255, 255, 255, 255)
  if gCEASEFIRE>0 then
    love.graphics.draw(iICON[12], 460, 280)
    drawshadowtext(math.floor(gCEASEFIRE/2), 460, 280, 60, "center", fTITLE, 150, 150, 150, 255)
    if mx>460 and mx<520 and my>280 and my<320 and gZOOMED==false then
      if gCEASEFIRE==1 then drawshadowtext("Last turn of ceasefire", mx+25, my, 150, "left", fDEFAULT, 255, 255, 255, 255, "Yes", "Yes", "Yes")
      elseif gCEASEFIRE<=3 then drawshadowtext("1 turn of ceasefire remaining", mx+25, my, 150, "left", fDEFAULT, 255, 255, 255, 255, "Yes", "Yes", "Yes")
      else drawshadowtext(math.floor(gCEASEFIRE/2).." turns of ceasefire remaining", mx+25, my, 150, "left", fDEFAULT, 255, 255, 255, 255, "Yes", "Yes", "Yes") end
    end
  end
  drawshadowtext("Player 1: "..tPOINTS[1].."/"..tBUDGET[1], 125, 558, 160, "left", fSUBHEADING, tCOLOUR[tSETTINGS.P1COLOUR].R, tCOLOUR[tSETTINGS.P1COLOUR].G, tCOLOUR[tSETTINGS.P1COLOUR].B, 255)
  drawshadowtext("Player 2: "..tPOINTS[2].."/"..tBUDGET[2], 125, 22, 160, "left", fSUBHEADING, tCOLOUR[tSETTINGS.P2COLOUR].R, tCOLOUR[tSETTINGS.P2COLOUR].G, tCOLOUR[tSETTINGS.P2COLOUR].B, 255)
  if tCONTROL[gTURN]=="Player" then
    if gACTION~="" then drawshadowtext("Cancel Move", xBUTTONS[gTURN]-25, yBUTTONS[gTURN]-45, 50, "center", fDEFAULT, 255, 215, 0, 255, "Yes", "Yes", "Yes", gFLASH)
    else drawshadowtext("Cancel Move", xBUTTONS[gTURN]-25, yBUTTONS[gTURN]-45, 50, "center", fDEFAULT, 60, 60, 60, 255, "No", "Yes", "Yes") end
    drawshadowtext("End Turn", xBUTTONS[gTURN]-25, yBUTTONS[gTURN], 50, "center", fDEFAULT, 0, 255, 0, 255, "Yes", "Yes", "Yes", gFLASH)
    drawshadowtext("Quit Game", xBUTTONS[gTURN]-25, yBUTTONS[gTURN]+45, 50, "center", fDEFAULT, 255, 0, 0, 255, "Yes", "Yes", "Yes", gFLASH)
  end
end

function drawselectedcard()
  local mx, my = love.mouse.getX() / gRATIOX, love.mouse.getY() / gRATIOY
  local i, r, moves
  if gZOOMED == true and gDRAGGING == false then
    drawcard(tCLICK[1].Card, 510, 300, 300, 420)
    if tCONTROL[gTURN]=="Player" then
      moves = calculatemoves(tCLICK[1].Player)
      if moves[4]==true then drawshadowtext("Target", 665, 110, 130, "center", fHEADING, 255, 255, 255, 255, "Yes", "Yes", "Yes", gFLASH)
      elseif moves[1]==true then drawshadowtext("Play to Combat", 665, 110, 130, "center", fHEADING, 255, 255, 255, 255, "Yes", "Yes", "Yes", gFLASH)
      elseif moves[2]==true then drawshadowtext("Play Action", 665, 110, 130, "center", fHEADING, 255, 255, 255, 255, "Yes", "Yes", "Yes", gFLASH)
      elseif moves[3]==true then drawshadowtext("Move / Attack", 665, 110, 130, "center", fHEADING, 255, 255, 255, 255, "Yes", "Yes", "Yes", gFLASH)
      elseif tCLICK[1].Source=="Hand" and tCLICK[1].Player == gTURN and gACTION == "" and tPOINTS[tCLICK[1].Player]<tCLICK[1].Card.Tier then drawshadowtext("Cannot afford to play card", 665, 110, 130, "center", fDEFAULT, 60, 60, 60, 255, "No", "Yes", "Yes")
      elseif tCLICK[1].Source=="Hand" and tCLICK[1].Player == gTURN and gACTION == "" and tPOINTS[tCLICK[1].Player]>=tCLICK[1].Card.Tier and (tCLICK[1].Card.Type=="Ally" or tCLICK[1].Card.Type=="Hero") and not getishomefree(tCLICK[1].Player) then drawshadowtext("Clear Home space to play card", 665, 110, 130, "center", fDEFAULT, 60, 60, 60, 255, "No", "Yes", "Yes")
      elseif tCLICK[1].Source=="Combat" and getreach(tCLICK[1].Card)==0 then drawshadowtext("Wait until next turn to move", 665, 110, 130, "center", fDEFAULT, 60, 60, 60, 255, "No", "Yes", "Yes")
      else end
      if moves[9]==true then drawshadowtext("Sacrifice Card", 665, 425, 130, "center", fHEADING, 255, 215, 0, 255, "Yes", "Yes", "Yes", gFLASH)
      elseif tCLICK[1].Source=="Hand" and gSACRIFICED == true then drawshadowtext("Sacrifice once per turn", 665, 425, 130, "center", fDEFAULT, 60, 60, 60, 255, "No", "Yes", "Yes")
      else end
      for i = 1, 4, 1 do
        if tCLICK[1].Card["Spec"..i]~=nil then
          if tCLICK[1].Source=="Combat" and tCLICK[1].Player == gTURN and gACTION == "" then
            r = getrechargetime(tCLICK[1].Card["Spec"..i])
            if moves[4+i]==true then drawshadowtext(tCLICK[1].Card["Spec"..i], 665, 125+(60*i), 130, "center", fDEFAULT, 0, 255, 255, 255, "Yes", "Yes", "Yes", gFLASH)
            elseif r==0 then drawshadowtext(tCLICK[1].Card["Spec"..i].." (automatic)", 665, 125+(60*i), 130, "center", fDEFAULT, 0, 100, 100, 255, "No", "Yes", "Yes")
            elseif tCLICK[1].Card["Charge"..i]<r then drawshadowtext(tCLICK[1].Card["Spec"..i].." (wait "..r-tCLICK[1].Card["Charge"..i]..")", 665, 125+(60*i), 130, "center", fDEFAULT, 0, 100, 100, 255, "No", "Yes", "Yes")
            else end
          end
        end
      end
    end
    for i = 1, 4, 1 do
      if tCLICK[1].Card["Spec"..i]~=nil then
        if mx>365 and mx<655 and my>340+(i*20) and my<360+(i*20) then drawshadowtext(getabilitydesc(tCLICK[1].Card["Spec"..i]), mx+25, my, 150, "left", fDEFAULT, 255, 255, 255, 255, "Yes", "Yes", "Yes") end
      end
    end
    r = 0
    for i = 1, #tEFFECTS, 1 do
      if tCLICK[1].Card.Effect~=nil and tCLICK[1].Card.Effect[i]>0 then
        r = r + 1
        if r<=3 and mx>377 and mx<417 and my>115+(r*45) and my<155+(r*45) then drawshadowtext(tEFFECTS[i].Name.."\n"..tEFFECTS[i].Result, mx+25, my, 150, "left", fDEFAULT, 255, 255, 255, 255, "Yes", "Yes", "Yes") 
        elseif r>3 and r<=6 and mx>605 and my<645 and my>115+((r-3)*45) and my<155+((r-3)*45) then drawshadowtext(tEFFECTS[i].Name.."\n"..tEFFECTS[i].Result, mx+25, my, 150, "left", fDEFAULT, 255, 255, 255, 255, "Yes", "Yes", "Yes") 
        else end
      end
    end
  end
end

function drawheldcard()
  local mx, my = love.mouse.getX() / gRATIOX, love.mouse.getY() / gRATIOY
  if gDRAGGING == true and gMOUSEGRAB > (tSETTINGS.FPS*0.3) and next(tCLICK)~=nil then drawcard(tCLICK[1].Card, mx, my, 50, 70, 0, 100) end
end

function drawendscreen()
  local mx, my = love.mouse.getX()/gRATIOX, love.mouse.getY()/gRATIOY
  local i
  if gBONUSCARD~=0 then
    drawshadowtext("You received a new card!", 50, 30, 700, "center", fHEADING, 255, 215, 0, 255)
    drawcard(tBLUEPRINT[gBONUSCARD], 400, 300, 300, 420)
    for i = 1, 4, 1 do
      if mx>255 and mx<545 and my>340+(i*20) and my<360+(i*20) and tBLUEPRINT[gBONUSCARD]["Spec"..i]~=nil then drawshadowtext(getabilitydesc(tBLUEPRINT[gBONUSCARD]["Spec"..i]), mx+25, my, 150, "left", fDEFAULT, 255, 255, 255, 255, "No", "Yes", "Yes") end
    end
  else drawshadowtext("You have unlocked all the cards for this level. Try a different level.", 50, 200, 700, "center", fHEADING, 255, 215, 0, 255) end
  if gNEWLEVEL ~=0 then drawshadowtext("You have unlocked the next level!", 50, 530, 700, "center", fHEADING, 255, 215, 0, 255) end
end

function draweditdeck()
  local collection = getlistofcards(tCOLLECTION, gTYPE, gTIER, gSORTBY)
  local deck = getlistofcards(tDECK[gTURN], gTYPE, gTIER, gSORTBY)
  local unique = {}
  local mx, my = love.mouse.getX()/gRATIOX, love.mouse.getY()/gRATIOY
  local i, j, found, card, row, col
  if next(collection)~=nil then
    for i = 1, #collection, 1 do
      found = false
      if next(unique)~=nil then
        for j = 1, #unique, 1 do
          if unique[j].Name==collection[i].Name then found = true end
        end
      end
      if found == false then table.insert(unique, deepcopy(collection[i])) end
    end
  end
  love.graphics.setColor(0, 0, 0, 200)
  love.graphics.rectangle("fill", 0, 0, 800, 600)
  love.graphics.setColor(100, 100, 100, 255)
  love.graphics.rectangle("fill", 0, 0, 800, 39)
  drawshadowtext("Type:", 0, 9, 58, "right", fSUBHEADING)
  drawshadowtext(gTYPE, 70, 9, 88, "center", fSUBHEADING, 255, 255, 255, 255, "Yes", "Yes", "Yes", gFLASH)
  drawshadowtext("Tier:", 170, 9, 53, "right", fSUBHEADING)
  if gTIER==0 then drawshadowtext("All", 235, 9, 88, "center", fSUBHEADING, 255, 255, 255, 255, "Yes", "Yes", "Yes", gFLASH)
  else drawshadowtext(gTIER, 235, 9, 88, "center", fSUBHEADING, 255, 255, 255, 255, "Yes", "Yes", "Yes", gFLASH) end
  drawshadowtext("Sort:", 330, 9, 58, "right", fSUBHEADING)
  drawshadowtext(gSORTBY, 400, 9, 88, "center", fSUBHEADING, 255, 255, 255, 255, "Yes", "Yes", "Yes", gFLASH)
  drawshadowtext("Load", 505, 9, 88, "center", fSUBHEADING, 255, 0, 0, 255, "Yes", "Yes", "Yes", gFLASH)
  drawshadowtext("Save", 605, 9, 88, "center", fSUBHEADING, 255, 215, 0, 255, "Yes", "Yes", "Yes", gFLASH)
  drawshadowtext("Accept", 705, 9, 88, "center", fSUBHEADING, 0, 255, 0, 255, "Yes", "Yes", "Yes", gFLASH)
  love.graphics.setColor(255, 215, 0, 255)
  love.graphics.line(497, 80, 497, 520)
  if gSELECT==1 and gSUBSELECT~=0 then card = unique[gSUBSELECT]
  elseif gSELECT==2 and gSUBSELECT~=0 then card = deck[gSUBSELECT]
  else end
  love.graphics.setColor(255, 255, 255, 255)
  drawshadowtext("All Cards", 5, 45, 486, "center", fSUBHEADING)
  drawshadowtext("Current Deck("..#tDECK[gTURN]..")", 503, 45, 292, "center", fSUBHEADING)
  love.graphics.setFont(fDEFAULT)
  if next(unique)~=nil then
    if gLINE1 > 1 then love.graphics.draw(iICON[3], 446, 30) end
    if gLINE1 < math.ceil(#unique/5)-3 then love.graphics.draw(iICON[4], 0, 30) end
    row, col, j = 1, 1, 1
    for i = 1, #unique, 1 do
      if row>=gLINE1 and row<=gLINE1+3 then 
        if getnumberinlist(unique[i].Name, tDECK[gTURN])<getnumberinlist(unique[i].Name, tCOLLECTION) then love.graphics.setColor(255, 255, 255, 255)
        else love.graphics.setColor(100, 100, 100, 255) end
        love.graphics.draw(iIMAGE[unique[i].Pic], (col*100)-50, (j*120), 0, 0.5, 0.5, 80, 80)
        love.graphics.printf(unique[i].Name, (col*100)-90, (j*120)+42, 80, "center")
        love.graphics.setColor(255, 255, 255, 255)
        if getnumberinlist(unique[i].Name,tDECK[gTURN])>0 then
          love.graphics.setColor(255, 255, 255, 255)
          love.graphics.rectangle("fill", (col*100)-95, (j*120)-10, 20, 20)
          love.graphics.setColor(0, 0, 0, 255)
          love.graphics.draw(iICON[6], (col*100)-85, (j*120), 0, 0.5, 0.5, 20, 20)
        end
        if getnumberinlist(unique[i].Name, tDECK[gTURN])<getnumberinlist(unique[i].Name, tCOLLECTION) then
          love.graphics.setColor(255, 255, 255, 255)
          love.graphics.rectangle("fill", (col*100)-25, (j*120)-10, 20, 20)
          love.graphics.setColor(0, 0, 0, 255)
          love.graphics.draw(iICON[5], (col*100)-15, (j*120), 0, 0.5, 0.5, 20, 20)
        end
      end
      col = col + 1
      if col > 5 then 
        if row>=gLINE1 then j = j + 1 end
        col = 1
        row = row + 1
      end
    end
  end
  if next(deck)~=nil then
    love.graphics.setColor(255, 255, 255, 255)
    if gLINE2 > 1 then love.graphics.draw(iICON[3], 750, 30) end
    if gLINE2 < math.ceil(#deck/3)-3 then love.graphics.draw(iICON[4], 498, 30) end
    row, col, j = 1, 1, 1
    for i = 1, #deck, 1 do
      if row>=gLINE2 and row<=gLINE2+3 then
        love.graphics.setColor(255, 255, 255, 255)
        love.graphics.draw(iIMAGE[deck[i].Pic], (col*100)+450, (j*120), 0, 0.5, 0.5, 80, 80)
        love.graphics.printf(deck[i].Name, (col*100)+410, (j*120)+42, 80, "center")
        love.graphics.setColor(0, 0, 0, 255)
        love.graphics.draw(iICON[17], (col*100)+483, (j*120)+32, 0, 0.5, 0.5, 15, 16)
        love.graphics.setColor(255, 255, 255, 255)
        love.graphics.draw(iICON[17], (col*100)+482, (j*120)+31, 0, 0.5, 0.5, 15, 16)
      end
      col = col + 1
      if col > 3 then 
        if row>=gLINE2 then j = j + 1 end
        col = 1
        row = row + 1
      end
    end
  end
  if gSELECT~=0 and gSUBSELECT~=0 then
    love.graphics.setColor(255, 255, 255, 100)
    love.graphics.rectangle("fill", 0, 0, 800, 600)
    drawcard(card, 400, 300, 300, 420)
    for i = 1, 4, 1 do
      if mx>255 and mx<545 and my>340+(i*20) and my<360+(i*20) and card["Spec"..i]~=nil then drawshadowtext(getabilitydesc(card["Spec"..i]), mx+25, my, 150, "left", fDEFAULT, 255, 255, 255, 255, "No", "Yes", "Yes") end
    end
    if gSELECT==1 then 
      if getnumberinlist(card.Name,tDECK[gTURN])>0 then drawshadowtext("Remove", 255, 518, 138, "center", fSUBHEADING, 255, 0, 0, 255, "Yes", "Yes", "Yes") end
      if getnumberinlist(card.Name,tDECK[gTURN])<getnumberinlist(card.Name, tCOLLECTION) then drawshadowtext("Add to Deck", 405, 518, 138, "center", fSUBHEADING, 0, 255, 0, 255, "Yes", "Yes", "Yes") end
    else drawshadowtext("Remove", 255, 518, 288, "center", fSUBHEADING, 255, 0, 0, 255, "Yes", "Yes", "Yes") end
  end
end

function drawsave()
  local i
  drawfiles()
  for i = 1, 26, 1 do
    drawshadowtext(string.char(i+96), 650+(i*25)-(math.ceil(i/7)*175), (math.ceil(i/7)*25)+25, 25, "center", fSUBHEADING, 255, 255, 255, 255, "Yes", "No", "No", gFLASH)
  end
  if gFILENAME ~= "" then drawshadowtext("Save", 500, 200, 200, "center", fSUBHEADING, 255, 215, 0, 255, "Yes", "Yes", "Yes", gFLASH) end
  love.graphics.setColor(255, 255, 255, 255)
  love.graphics.draw(iICON[7], 625, 125)
end

function drawload()
  drawfiles()
  if gFILENAME ~= "" then 
    drawshadowtext("Load", 500, 200, 200, "center", fSUBHEADING, 0, 255, 0, 255, "Yes", "Yes", "Yes", gFLASH)
    drawshadowtext("Erase", 500, 240, 200, "center", fSUBHEADING, 255, 0, 0, 255, "Yes", "Yes", "Yes", gFLASH)
  end
end

function drawfiles()
  local list = getordereddirectory("decks")
  local i
  drawshadowtext("Decks", 0, 10, 400, "center", fHEADING, 255, 215, 0, 255)
  if next(list)~=nil then
    for i = #list, 1, -1 do
      if string.sub(list[i], -4)~=".txt" then table.remove(list, i) end
    end
  end
  if next(list)~=nil then
    for i = 1, #list, 1 do
      if i<=25 then drawshadowtext(list[i], 20, (i*20)+30, 360, "left", fDEFAULT, 255, 255, 255, 255, "Yes", "No", "No", gFLASH) end
    end
  end
  drawshadowtext(gFILENAME, 500, 160, 200, "center", fSUBHEADING)
  drawshadowtext("Back", 700, 8, 88, "center", fSUBHEADING, 255, 255, 255, 255, "Yes", "Yes", "Yes", gFLASH)
end

function love.mousepressed(x, y, button, istouch, presses)  -- Record any click event and process based on gamemode
  x, y = x/gRATIOX, y/gRATIOY
  playsound(1)
  if _G["click"..gGAMEMODE]~=nil then _G["click"..gGAMEMODE](x, y, presses) end   -- Click handlers based on gamemode
end

function clicktitle(x, y, p)  -- Click title page
  gFADE = 255
  gGAMEMODE = "menu"
  iBACKGROUND = love.graphics.newImage("backgrounds/background12.jpg")
  playtrack(2)
end

function clickmenu(x, y, p)  -- Click main menu
  if x>300 and x<500 and y>200 and y<232 then 
    gGAMEMODE = "map"
    tCONTROL = {"Player", "AI"}
    tDECK[1] = loaddeck("deck.txt")
    gSELECT = tPLAYER.LEVEL
    gSUBSELECT = 0
  elseif x>300 and x<500 and y>250 and y<282 then
    tCONTROL = {"Player", "Player"}
    gLEVEL, gDIFFICULTY = 0, 0
    addpopup("P1: Choose Deck")
    tDECK[1] = loaddeck("deck.txt")
    tDECK[2] = loaddeck("deck.txt")
    openeditdeck()
  elseif x>300 and x<500 and y>300 and y<332 then
    tCONTROL = {"Player", "AI"}
    tDECK[1] = loaddeck("deck.txt")
    openeditdeck()
  elseif x>300 and x<500 and y>350 and y<382 then gGAMEMODE = "settings"
  elseif x>300 and x<500 and y>400 and y<432 then quitgame()
  else end
end

function openeditdeck()
  gGAMEMODE = "editdeck"
  gTURN = 1
  gTYPE = "All"
  gTIER = 0
  gSORTBY = "Type"
  gSELECT = 0
  gSUBSELECT = 0
  gLINE1 = 1
  gLINE2 = 1
end

function clicksettings(x, y, p)  -- Click settings menu
  local list = {"FPS", "Sound", "Music", "MusicVol", "Gridsize", "P1COLOUR", "P2COLOUR"}
  local i
  for i = 1, #list, 1 do
    if (x>210 and x<400 and y>100+(i*40) and y<132+(i*40)) or (x>610 and x<800 and y>100+((i-8)*40) and y<132+((i-8)*40)) then
      if list[i]=="FPS" then tSETTINGS.FPS = getnextoption(tSETTINGS.FPS, {30, 40, 50, 60, 70, 80, 90})
      elseif list[i]=="Sound" then 
        tSETTINGS.Sound = getnextoption(tSETTINGS.Sound, {"Yes", "No"})
        playsound(1)
      elseif list[i]=="Music" then 
        tSETTINGS.Music = getnextoption(tSETTINGS.Music, {"Yes", "No"})
        if tSETTINGS.Music == "No" then musicstop()
        else playtrack(2) end
      elseif list[i]=="MusicVol" then 
        tSETTINGS.MusicVol = getnextoption(tSETTINGS.MusicVol, {0.10, 0.25, 0.50, 0.75, 1.00})
        musicsetvolume(tSETTINGS.MusicVol)
      elseif list[i]=="Gridsize" then tSETTINGS.Gridsize = getnextoption(tSETTINGS.Gridsize, {"Small", "Medium", "Large"})
      elseif list[i]=="P1COLOUR" then
        tSETTINGS.P1COLOUR = tSETTINGS.P1COLOUR + 1
        if tSETTINGS.P1COLOUR > #tCOLOUR then tSETTINGS.P1COLOUR = 1 end
      elseif list[i]=="P2COLOUR" then
        tSETTINGS.P2COLOUR = tSETTINGS.P2COLOUR + 1
        if tSETTINGS.P2COLOUR > #tCOLOUR then tSETTINGS.P2COLOUR = 1 end
      else end
    end
  end
  if x>300 and x<500 and y>475 and y<507 then
    writesingletable("settings.txt", tSETTINGS)
    gGAMEMODE = "menu"
  end
end

function clickmap(x, y, p)  -- Click campaign map
  local click = false
  local i
  if x>720 and x<790 and y>560 and y<595 then gGAMEMODE = "menu"
  else
    for i = 1, tPLAYER.LEVEL, 1 do
      if gSELECT == i and gSUBSELECT == i then
        if x>tMAP[i].X+20 and x<tMAP[i].X+70 and y>tMAP[i].Y-36 and y<tMAP[i].Y+34 then
          if y<tMAP[i].Y-13 then gSUBSELECT = 1
          elseif y>tMAP[i].Y+11 then gSUBSELECT = 3
          else gSUBSELECT = 2 end
          gLEVEL, gDIFFICULTY = gSELECT, gSUBSELECT
          tDECK[2] = loaddeck("ai/deck"..tostring((3*(gSELECT-1))+gSUBSELECT)..".txt")
          tPLAYER.GAMES = tPLAYER.GAMES + 1
          newgame()
        end
      elseif distancebetween(x, y, tMAP[i].X, tMAP[i].Y)<=50 then
        click = true
        if gSELECT ~= i then 
          gSELECT = i
          gSUBSELECT = 0
        elseif gSUBSELECT ~= i then gSUBSELECT = i
        else end
      else end
    end
    if click == false then gSELECT, gSUBSELECT = 0, 0 end
  end
end

function clickgame(x, y, p)   -- Click game
  if gZOOMED == false and gENDING == 0 then clicksurface(x, y, p)
  elseif gZOOMED == true and gENDING == 0 then clickselectedcard(x, y, p)
  else end
end

function clicksurface(x, y, p)   -- Click in-game, not on the zoom selected view
  local clicked = false
  local i, j
  if x>xBUTTONS[gTURN]-25 and x<xBUTTONS[gTURN]+25 and y>yBUTTONS[gTURN]-45 and y<yBUTTONS[gTURN] and gACTION~="" and clicked == false then
    tCLICK = {}
    gACTION = ""
    clicked = true
  elseif x>xBUTTONS[gTURN]-25 and x<xBUTTONS[gTURN]+25 and y>yBUTTONS[gTURN] and y<yBUTTONS[gTURN]+45 and clicked == false then
    gTURN = 3 - gTURN
    newturn(gTURN)
    clicked = true
  elseif x>xBUTTONS[gTURN]-25 and x<xBUTTONS[gTURN]+25 and y>yBUTTONS[gTURN]+45 and y<yBUTTONS[gTURN]+90 and clicked == false then
    gGAMEMODE = "menu"
    iBACKGROUND = love.graphics.newImage("backgrounds/background12.jpg")
    gFADE = 255
    playtrack(2)
    clicked = true
  else end
  for j = 1, 2, 1 do
    if next(tDECK[j])~=nil then
      if tDECK[j][1].Hover==true and clicked == false then
        tDECK[j][1].Hover=false
        clickcard(tDECK[j][1], "Deck", j, 1)
        clicked = true
      end
    end
    if next(tHAND[j])~=nil then
      for i = 1, #tHAND[j], 1 do
        if tHAND[j][i].Hover==true and clicked == false then
          tHAND[j][i].Hover=false
          clickcard(tHAND[j][i], "Hand", j, i)
          clicked = true
        end
      end
    end
    if next(tGRAVE[j])~=nil then
      if tGRAVE[j][#tGRAVE[j]].Hover==true and clicked == false then
        tGRAVE[j][#tGRAVE[j]].Hover=false
        clickcard(tGRAVE[j][#tGRAVE[j]], "Grave", j, #tGRAVE[j])
        clicked = true
      end
    end
    if next(tCOMBAT[j])~=nil then
      for i = 1, #tCOMBAT[j], 1 do
        if tCOMBAT[j][i].Hover==true and clicked == false then
          tCOMBAT[j][i].Hover=false
          clickcard(tCOMBAT[j][i], "Combat", j, i)
          clicked = true
        end
      end
    end
  end
  if x>xGRID-(2.5*tSETTINGS.HexWidth) and x<xGRID+(2.5*tSETTINGS.HexWidth) and y>yGRID-(3.5*tSETTINGS.HexHeight) and y<yGRID+(3.5*tSETTINGS.HexHeight) and clicked == false then 
    i, j = screentogrid(x, y)
    if i>=1 and i<=gGRID_WIDTH and j>=1 and j<=gGRID_HEIGHT then
      if tHEX[j][i].Player==0 and gACTION=="Move/Attack" and next(tCLICK)~=nil and tHEX[j][i].Moves<=getreach(tCLICK[#tCLICK].Card) then domove(i, j)
      elseif tHEX[j][i].Player==0 and gACTION=="Charge" and next(tCLICK)~=nil and tHEX[j][i].Moves<=getreach(tCLICK[#tCLICK].Card) then docharge(i, j)
      else end
    end
  end
end

function clickcard(card, source, player, index)
  local click = {}
  click.Card = card
  click.Source = source
  click.Player = player
  click.Index = index
  table.insert(tCLICK, 1, deepcopy(click))
  gDRAGGING = true
  gZOOMED = true
end

function clickselectedcard(x, y, p)
  local clear = false
  local card, moves, targets
  if next(tCLICK)~=nil then
    if tCLICK[1].Source == "Hand" then card = tHAND[tCLICK[1].Player][tCLICK[1].Index]
    elseif tCLICK[1].Source == "Deck" then card = tDECK[tCLICK[1].Player][tCLICK[1].Index]
    elseif tCLICK[1].Source == "Grave" then card = tGRAVE[tCLICK[1].Player][tCLICK[1].Index]
    elseif tCLICK[1].Source == "Combat" then card = tCOMBAT[tCLICK[1].Player][tCLICK[1].Index]
    else end
  end
  if gZOOMED == true and gDRAGGING == false then
    moves = calculatemoves(gTURN)
    if moves[4]==true and tCONTROL[gTURN]=="Player" and x>665 and x<795 and y>110 and y<175 then   -- Target this card
      if gACTION == "Move/Attack" or #tCLICK-1 == getabilitytargets(gACTION) then clear = true end
      dotarget()
      if clear == true then
        gACTION = ""
        tCLICK = {}
        cleartargets()
      end
      gZOOMED = false
    elseif moves[1]==true and x>665 and x<795 and y>110 and y<175 then       -- March ally
      tPOINTS[gTURN] = tPOINTS[gTURN] - tCLICK[1].Card.Tier
      addpopup("P"..gTURN.." plays "..card.Name.." to combat")
      sendcard("Combat", card, tCLICK[1].Source, tCLICK[1].Player, tCLICK[1].Index)
      tCLICK = {}
      gZOOMED = false
    elseif moves[2]==true and x>665 and x<795 and y>110 and y<175 then       -- Play action
      gACTION = card.Spec1
      if playaction(card, gTURN) == true then
        tPOINTS[gTURN] = tPOINTS[gTURN] - tCLICK[1].Card.Tier
        targets = getabilitytargets(gACTION)
        sendcard("Grave", card, tCLICK[1].Source, tCLICK[1].Player, tCLICK[1].Index)
        if targets == 0 and next(tCLICK)~=nil then
          table.remove(tCLICK,1)
          gACTION = ""
        end
      else
        gACTION = ""
        tCLICK = {}
      end
      gZOOMED = false
    elseif moves[3]==true and x>665 and x<795 and y>110 and y<175 then       -- Move/Attack
      gACTION = "Move/Attack"
      refreshmoves()
      if card.Effect[4]>0 then addpopup(card.Name.." is scared and can only move")
      elseif findtargets() then addpopup("Choose target")
      else addpopup("Choose location") end
      gZOOMED = false
    elseif moves[5]==true and x>665 and x<795 and y>185 and y<235 then        -- Use special ability 1
      gACTION = card.Spec1
      if playaction(card, gTURN) == true then
        card.Charge1 = 0
        if getabilitytargets(gACTION) == 0 and next(tCLICK)~=nil and gACTION~="Charge" then
          table.remove(tCLICK,1)
          gACTION = ""
        end
      else 
        gACTION = ""
        tCLICK = {}
      end
      gZOOMED = false
    elseif moves[6]==true and x>665 and x<795 and y>245 and y<295 then        -- Use special ability 2
      gACTION = card.Spec2
      if playaction(card, gTURN) == true then
        card.Charge2 = 0
        if getabilitytargets(gACTION) == 0 and next(tCLICK)~=nil and gACTION~="Charge" then
          table.remove(tCLICK,1)
          gACTION = ""
        end
      else 
        gACTION = ""
        tCLICK = {}
      end
      gZOOMED = false
    elseif moves[7]==true and x>665 and x<795 and y>305 and y<355 then        -- Use special ability 3
      gACTION = card.Spec3
      if playaction(card, gTURN) == true then
        card.Charge3 = 0
        if getabilitytargets(gACTION) == 0 and next(tCLICK)~=nil and gACTION~="Charge" then
          table.remove(tCLICK,1)
          gACTION = ""
        end
      else 
        gACTION = ""
        tCLICK = {}
      end
      gZOOMED = false
    elseif moves[8]==true and x>665 and x<795 and y>365 and y<415 then        -- Use special ability 4
      gACTION = card.Spec4
      if playaction(card, gTURN) == true then
        card.Charge4 = 0
        if getabilitytargets(gACTION) == 0 and next(tCLICK)~=nil and gACTION~="Charge" then
          table.remove(tCLICK,1)
          gACTION = ""
        end
      else 
        gACTION = ""
        tCLICK = {}
      end
      gZOOMED = false
    elseif moves[9]==true and x>665 and x<795 and y>425 and y<490 then        -- Sacrifice card
      addpopup("P"..gTURN.." sacrifices "..card.Name)
      sendcard("Grave", card, tCLICK[1].Source, tCLICK[1].Player, tCLICK[1].Index)
      tBUDGET[gTURN] = tBUDGET[gTURN] + 1
      tPOINTS[gTURN] = tPOINTS[gTURN] + 1 
      gSACRIFICED = true
      tCLICK = {}
      gZOOMED = false
    elseif x>360 and x<660 and y>90 and y<510 then                      -- Do nothing, clicked the zoomed in card

    else                                                                -- Zoom out and remove this card from tCLICK
      gZOOMED = false
      if next(tCLICK)~=nil then table.remove(tCLICK,1) end
    end
  end
end

function clickendscreen(x, y, p)
  gFADE = 255
  iBACKGROUND = love.graphics.newImage("backgrounds/background12.jpg")
  gGAMEMODE = "menu"
  playtrack(2)
end

function clickeditdeck(x, y, p)
  local collection = getlistofcards(tCOLLECTION, gTYPE, gTIER, gSORTBY)
  local deck = getlistofcards(tDECK[gTURN], gTYPE, gTIER, gSORTBY)
  local unique = {}
  local i, j, found, row, col
  if next(collection)~=nil then
    for i = 1, #collection, 1 do
      found = false
      if next(unique)~=nil then
        for j = 1, #unique, 1 do
          if unique[j].Name==collection[i].Name then found = true end
        end
      end
      if found == false then table.insert(unique, deepcopy(collection[i])) end
    end
  end
  if gSELECT~=0 and gSUBSELECT~=0 then 
    if gSELECT==1 and x>=255 and x<=393 and y>=518 and y<=541 and getnumberinlist(unique[gSUBSELECT].Name, tDECK[gTURN])>0 then 
      removecardfromdeck(unique[gSUBSELECT].Name, gTURN)
    elseif gSELECT==1 and x>=405 and x<=543 and y>=518 and y<=541 and getnumberinlist(unique[gSUBSELECT].Name, tDECK[gTURN])<getnumberinlist(unique[gSUBSELECT].Name, tCOLLECTION) then 
      table.insert(tDECK[gTURN], deepcopy(unique[gSUBSELECT]))
    elseif gSELECT==2 and x>=255 and x<=543 and y>=518 and y<=541 and getnumberinlist(deck[gSUBSELECT].Name, tDECK[gTURN])>0 then 
      removecardfromdeck(deck[gSUBSELECT].Name, gTURN)
      gSELECT, gSUBSELECT = 0, 0
    elseif x>=250 and x<=450 and y>=90 and y<=510 then -- Do nothing, clicked close up card
    else gSELECT, gSUBSELECT = 0, 0 end
  elseif x>705 and x<793 and y>9 and y<30 then
    if tCONTROL[2]=="AI" then
      savedeck(tDECK[gTURN], "deck.txt")
      gGAMEMODE="menu"
    elseif gTURN == 1 then
      addpopup("P2: Choose Deck")
      gTURN=2
    else
      gLEVEL, gDIFFICULTY = 0, 0
      newgame()
    end
  elseif x>605 and x<692 and y>9 and y<30 then
    gFILENAME = ""
    gGAMEMODE = "save"
    addpopup("Choose filename")
  elseif x>505 and x<592 and y>9 and y<30 then
    gFILENAME = ""
    gGAMEMODE = "load"
    addpopup("Choose filename")
  elseif x>70 and x<158 and y>9 and y<30 then 
    gTYPE = getnextoption(gTYPE, {"All", "Hero", "Ally", "Action"})
    gLINE1, gLINE2 = 1, 1
  elseif x>235 and x<323 and y>9 and y<30 then 
    gTIER = getnextoption(gTIER, {0, 1, 2, 3, 4, 5, 6, 7, 8, 9})
    gLINE1, gLINE2 = 1, 1
  elseif x>400 and x<488 and y>9 and y<30 then 
    gSORTBY = getnextoption(gSORTBY, {"Tier", "Type"})
    gLINE1, gLINE2 = 1, 1
  elseif x>10 and x<40 and y>40 and y<70 and gLINE1 < math.ceil(#unique/5)-3 then gLINE1 = gLINE1 + 1
  elseif x>456 and x<486 and y>40 and y<70 and gLINE1 > 1 then gLINE1 = gLINE1 - 1
  elseif x>508 and x<538 and y>40 and y<70 and gLINE2 < math.ceil(#deck/3)-3 then gLINE2 = gLINE2 + 1
  elseif x>760 and x<790 and y>40 and y<70 and gLINE2 > 1 then gLINE2 = gLINE2 - 1
  elseif x>0 and x<495 and y>=80 and y<=520 then
    gSELECT = 1
    if next(unique)~=nil then
      row, col, j = 1, 1, 1
      for i = 1, #unique, 1 do
        if row>=gLINE1 and row<=gLINE1+3 then 
          if x>=(col*100)-95 and x<=(col*100)-75 and y>=(j*120)-10 and y<=(j*120)+10 and getnumberinlist(unique[i].Name, tDECK[gTURN])>0 then removecardfromdeck(unique[i].Name, gTURN)
          elseif x>=(col*100)-25 and x<=(col*100)-5 and y>=(j*120)-10 and y<=(j*120)+10 and getnumberinlist(unique[i].Name, tDECK[gTURN])<getnumberinlist(unique[i].Name, tCOLLECTION) then table.insert(tDECK[gTURN], deepcopy(unique[i]))
          elseif x>=(col*100)-90 and x<=(col*100)-10 and y>=(j*120)-40 and y<=(j*120)+40 then gSUBSELECT = i
          else end
        end
        col = col + 1
        if col > 5 then
          if row>=gLINE1 then j = j + 1 end
          col = 1
          row = row + 1
        end
      end
    end
  elseif x>=510 and x<=790 and y>=80 and y<=520 then
    gSELECT = 2
    if next(deck)~=nil then
      row, col, j = 1, 1, 1
      for i = 1, #deck, 1 do
        if row>=gLINE2 and row<=gLINE2+3 then
          if x>=(col*100)+450 and x<=(col*100)+490 and y>=(j*120) and y<=(j*120)+40 then removecardfromdeck(deck[i].Name, gTURN)
          elseif x>=(col*100)+410 and x<=(col*100)+490 and y>=(j*120)-40 and y<=(j*120)+40 then gSUBSELECT = i
          else end
        end
        col = col + 1
        if col > 3 then
          if row>=gLINE2 then j = j + 1 end
          col = 1
          row = row + 1
        end
      end
    end
  else end
end

function removecardfromdeck(name, pl)
  local found = false
  local i
  if next(tDECK[pl])~=nil then
    for i = #tDECK[pl], 1, -1 do
      if tDECK[pl][i].Name==name and found==false then
        found = true
        table.remove(tDECK[pl], i)
      end
    end
  end
end

function clicknewturn(x, y, p)
  gGAMEMODE = "game"
end

function clickload(x, y, p)
  local list = getordereddirectory("decks")
  local i
  if next(list)~=nil then
    for i = #list, 1, -1 do
      if string.sub(list[i], -4)~=".txt" then table.remove(list, i) end
    end
  end
  if x>500 and x<700 and y>200 and y<232 and gFILENAME ~= "" then 
    tDECK[gTURN] = loaddeck("decks/"..gFILENAME)
    gGAMEMODE = "editdeck"
  elseif x>500 and x<700 and y>240 and y<272 and gFILENAME ~= "" then
    love.filesystem.remove("decks/"..gFILENAME)
    gFILENAME = ""
  elseif x>700 and x<788 and y>8 and y<40 then gGAMEMODE = "editdeck"
  elseif next(list)~=nil then
    for i = 1, #list, 1 do
      if x>20 and i<380 and y>(i*20)+30 and y<(i*20)+46 then gFILENAME = list[i] end
    end
  else end
end

function clicksave(x, y, p)
  local list = getordereddirectory("decks")
  local i
  if next(list)~=nil then
    for i = #list, 1, -1 do
      if string.sub(list[i], -4)~=".txt" then table.remove(list, i) end
    end
  end
  if x>500 and x<700 and y>200 and y<232 and gFILENAME ~= "" then 
    if string.sub(gFILENAME, -4)~=".txt" then gFILENAME = gFILENAME..".txt" end
    savedeck(tDECK[gTURN], "decks/"..gFILENAME)
    gGAMEMODE = "editdeck"
  elseif x>500 and x<700 and y>50 and y<150 then
    for i = 1, 26, 1 do
      if x>650+(i*25)-(math.ceil(i/7)*175) and x<675+(i*25)-(math.ceil(i/7)*175) and y>(math.ceil(i/7)*25)+25 and y<(math.ceil(i/7)*25)+50 then
        gFILENAME = gFILENAME..string.char(i+96)
        if string.len(gFILENAME)>20 then gFILENAME = string.sub(gFILENAME, 1, 20) end
      end
    end
    if x>625 and x<650 and y>125 and y<150 then
      if string.len(gFILENAME)>=2 then gFILENAME = string.sub(gFILENAME, 1, string.len(gFILENAME)-1)
      else gFILENAME = "" end
    end
  elseif x>700 and x<788 and y>8 and y<40 then gGAMEMODE = "editdeck"
  elseif next(list)~=nil then
    for i = 1, #list, 1 do
      if x>20 and i<380 and y>(i*20)+30 and y<(i*20)+46 then gFILENAME = list[i] end
    end
  else end
end

function love.mousereleased(x, y, button, istouch, presses) -- Mouse released
  x, y = x/gRATIOX, y/gRATIOY
  if _G["release"..gGAMEMODE]~=nil then _G["release"..gGAMEMODE](x, y) end   -- Release handler based on gamemode
  gMOUSEGRAB = 0
  gDRAGGING = false
end

function releasegame(x,y)
  local a, b = gTURN, 3-gTURN
  local card, moves, targets, click
  local i, gx, gy
  
  if next(tCLICK)~=nil then
    if tCLICK[1].Source == "Hand" then card = tHAND[tCLICK[1].Player][tCLICK[1].Index]
    elseif tCLICK[1].Source == "Deck" then card = tDECK[tCLICK[1].Player][tCLICK[1].Index]
    elseif tCLICK[1].Source == "Grave" then card = tGRAVE[tCLICK[1].Player][tCLICK[1].Index]
    elseif tCLICK[1].Source == "Combat" then card = tCOMBAT[tCLICK[1].Player][tCLICK[1].Index]
    else end
  end
  
  if gDRAGGING == true and next(tCLICK)~=nil and tCONTROL[gTURN]=="Player" then
    moves = calculatemoves(a)
    if moves[1]==true and x>xGRID-(gGRID_WIDTH*tSETTINGS.HexWidth/2) and x<xGRID+(gGRID_WIDTH*tSETTINGS.HexWidth/2) and y>yGRID-(gGRID_HEIGHT*tSETTINGS.HexHeight*0.375) and y<yGRID+(gGRID_HEIGHT*tSETTINGS.HexHeight*0.375) then    -- March ally
      tPOINTS[a] = tPOINTS[a] - card.Tier
      addpopup("P"..gTURN.." plays "..card.Name.." to combat")
      sendcard("Combat", card, tCLICK[1].Source, tCLICK[1].Player, tCLICK[1].Index)
      tCLICK = {}
      gZOOMED = false
    elseif moves[2]==true and x>xGRID-(gGRID_WIDTH*tSETTINGS.HexWidth/2) and x<xGRID+(gGRID_WIDTH*tSETTINGS.HexWidth/2) and y>yGRID-(gGRID_HEIGHT*tSETTINGS.HexHeight*0.375) and y<yGRID+(gGRID_HEIGHT*tSETTINGS.HexHeight*0.375) then    -- Play an action
      gACTION = card.Spec1
      if playaction(card, a) == true then
        tPOINTS[a] = tPOINTS[a] - card.Tier
        targets = getabilitytargets(gACTION)
        sendcard("Grave", card, tCLICK[1].Source, tCLICK[1].Player, tCLICK[1].Index)
        if targets == 0 and next(tCLICK)~=nil then -- Do nothing, dealt with below
        elseif findtargets() then
          if next(tGRAVE[a])~=nil then
            i = #tGRAVE[a]
            if tGRAVE[a][i].Hover == true and tGRAVE[a][i].Target == true then
              clickcard(tGRAVE[a][i], "Grave", a, i)
              dotarget()
              targets = targets - 1
            end
          end
          if next(tGRAVE[b])~=nil then
            i = #tGRAVE[b]
            if tGRAVE[b][i].Hover == true and tGRAVE[b][i].Target == true then
              clickcard(tGRAVE[b][i], "Grave", b, i)
              dotarget()
              targets = targets - 1
            end
          end
          if next(tDECK[a])~=nil then
            if tDECK[a][1].Hover == true and tDECK[a][1].Target == true then
              clickcard(tDECK[a][1], "Deck", a, 1)
              dotarget()
              targets = targets - 1
            end
          end
          if next(tDECK[b])~=nil then
            if tDECK[b][1].Hover == true and tDECK[b][1].Target == true then
              clickcard(tDECK[b][1], "Deck", b, 1)
              dotarget()
              targets = targets - 1
            end
          end
          if next(tHAND[a])~=nil then
            for i = #tHAND[a], 1, -1 do
              if tHAND[a][i].Hover == true and tHAND[a][i].Target == true then
                clickcard(tHAND[a][i], "Hand", a, i)
                dotarget()
                targets = targets - 1
              end
            end
          end
          if next(tHAND[b])~=nil then
            for i = #tHAND[b], 1, -1 do
              if tHAND[b][i].Hover == true and tHAND[b][i].Target == true then
                clickcard(tHAND[b][i], "Hand", b, i)
                dotarget()
                targets = targets - 1
              end
            end
          end
          if next(tCOMBAT[a])~=nil then
            for i = #tCOMBAT[a], 1, -1 do
              if tCOMBAT[a][i].Hover == true and tCOMBAT[a][i].Target == true then
                clickcard(tCOMBAT[a][i], "Combat", a, i)
                dotarget()
                targets = targets - 1
              end
            end
          end
          if next(tCOMBAT[b])~=nil then
            for i = #tCOMBAT[b], 1, -1 do
              if tCOMBAT[b][i].Hover == true and tCOMBAT[b][i].Target == true then
                clickcard(tCOMBAT[b][i], "Combat", b, i)
                dotarget()
                targets = targets - 1
              end
            end
          end
        else 
          addpopup("No valid targets")
          targets = 0
        end
        if targets == 0 then 
          tCLICK = {}
          gACTION = ""
          cleartargets()
        end
        gZOOMED = false
      else
        gACTION = ""
        tCLICK = {}
        gZOOMED = false
      end
    elseif moves[3]==true and x>xGRID-(gGRID_WIDTH*tSETTINGS.HexWidth/2) and x<xGRID+(gGRID_WIDTH*tSETTINGS.HexWidth/2) and y>yGRID-(gGRID_HEIGHT*tSETTINGS.HexHeight*0.375) and y<yGRID+(gGRID_HEIGHT*tSETTINGS.HexHeight*0.375) then    -- Move/Attack
      gACTION = "Move/Attack"
      refreshmoves()
      targets = false
      if findtargets() and card.Effect[4]==0 and next(tCOMBAT[b])~=nil then
        for i = #tCOMBAT[b], 1, -1 do
          if tCOMBAT[b][i].Hover == true and tCOMBAT[b][i].Target == true then
            clickcard(tCOMBAT[b][i], "Combat", b, i)
            dotarget()
            targets = true
          end
        end
      end
      if targets == false then
        gx, gy = screentogrid(x, y)
        if tHEX[gy][gx].Player==0 and tHEX[gy][gx].Moves<=getreach(card) then 
          domove(gx, gy) 
          targets = true
        end
      end
      gACTION = ""
      if targets == true then
        tCLICK = {}
        cleartargets()
        gZOOMED = false
      end
    elseif moves[9]==true and x>xGRAVE[gTURN]-(tSETTINGS.CardWidth/2) and x<xGRAVE[gTURN]+(tSETTINGS.CardWidth/2) and y>yGRAVE[gTURN]-(tSETTINGS.CardHeight/2) and y<yGRAVE[gTURN]+(tSETTINGS.CardHeight/2) then   -- Sacrifice card
      addpopup("P"..gTURN.." sacrifices "..card.Name)
      sendcard("Grave", card, tCLICK[1].Source, tCLICK[1].Player, tCLICK[1].Index)
      tBUDGET[gTURN] = tBUDGET[gTURN] + 1
      tPOINTS[gTURN] = tPOINTS[gTURN] + 1 
      gSACRIFICED = true
      tCLICK = {}
      gZOOMED = false
    else end
  end
end 

function playsound(i)  -- Play sound
  if sSOUND[i] ~= nil and tSETTINGS.Sound == "Yes" then 
    sSOUND[i]:stop()
    love.audio.play(sSOUND[i])
  end
end

function playtrack(i) -- Play music
  if tSETTINGS.Music == "Yes" then musicplay(i) end
end

function gethealth(card)
  local h = card.MaxHealth
  if card.Wounds~=nil then h = h - card.Wounds end
  return h
end

function getarmour(card)
  local arm = 0
  local i
  for i = 1, 5, 1 do
    if gethasability(card, "Armour +"..i) then arm = arm + i end
  end
  return arm
end

function getkilldmg(card)
  return gethealth(card) + getarmour(card)
end

function getpower(card)
  local p = card.Power
  if card.Effect~=nil and next(card.Effect)~=nil then
    if card.Effect[5]>0 then p = p * 1.25 end
    if card.Effect[6]>0 then p = p * 1.33 end
    if card.Effect[7]>0 then p = p * 0.66 end
    if card.Effect[9]>0 then p = p * 1.5 end
    if card.Effect[10]>0 then p = p * 0.5 end
    if card.Effect[14]>0 then p = p * 1.25 end
    if card.Effect[16]>0 then p = p + 1 end
    if card.Effect[17]>0 then p = p - 1 end
  end
  return math.floor(p+0.5)
end

function getreach(card)
  local r = card.Moves
  if card.Effect[15]>0 then r = r + 1 end
  r = r - card.Steps
  return r
end

function getrechargetime(ability)
  local t = 0
  local i
  if next(tABILITIES)~=nil then
    for i = 1, #tABILITIES, 1 do
      if tABILITIES[i].Name == ability then t = tABILITIES[i].Recharge end
    end
  end
  return t
end

function getabilitydesc(ability)
  local d = ""
  local i
  if next(tABILITIES)~=nil then
    for i = 1, #tABILITIES, 1 do
      if tABILITIES[i].Name == ability then d = tABILITIES[i].Desc end
    end
  end
  return d
end

function getabilitytargets(ability)
  local t = 0
  local i
  if next(tABILITIES)~=nil then
    for i = 1, #tABILITIES, 1 do
      if tABILITIES[i].Name == ability then t = tABILITIES[i].Targets end
    end
  end
  return t
end

function gethasability(card, ability)
  local test = false
  local i
  for i = 1, 4, 1 do
    if card["Spec"..i]~=nil then
      if card["Spec"..i]==ability then test = true end
    end
  end
  return test
end

function getnumberinlist(name, list)
  local count = 0
  local i
  if next(list)~=nil then
    for i = 1, #list, 1 do
      if list[i].Name==name then count = count + 1 end
    end
  end
  return count
end

function getlistofcards(list, kind, tier, sortby)
  local unsorted, sorted, selected = {}, {}, {}
  local i, j, score, best, ref
  kind = kind or "All"
  tier = tier or 0
  sortby = sortby or "Type"
  if next(list)~=nil then
    for i = 1, #list, 1 do
      if (kind == "All" or list[i].Type==kind) and (tier == 0 or list[i].Tier==tier) then table.insert(unsorted, deepcopy(list[i])) end
    end
  end
  if next(unsorted)~=nil then
    for j = 1, #unsorted, 1 do
      selected[j] = false
    end
    for j = 1, #selected, 1 do
      best, ref = 0, 0
      for i = 1, #unsorted, 1 do
        if selected[i]==false then
          if sortby == "Tier" then 
            score = (11-unsorted[i].Tier)*10
            if unsorted[i].Type == "Hero" then score = score + 3
            elseif unsorted[i].Type == "Ally" then score = score + 2
            elseif unsorted[i].Type == "Action" then score = score + 1
            else end
          else
            score = 11-unsorted[i].Tier
            if unsorted[i].Type == "Hero" then score = score + 30
            elseif unsorted[i].Type == "Ally" then score = score + 20
            elseif unsorted[i].Type == "Action" then score = score + 10
            else end
          end
          if score>best or (score==best and (ref==0 or unsorted[i].Name<unsorted[ref].Name)) then
            best = score
            ref = i
          end
        end
      end
      if best ~= 0 and ref ~= 0 then
        table.insert(sorted, deepcopy(unsorted[ref]))
        selected[ref] = true
      end
    end
  end
  return sorted
end

function getfreespaces(x, y)
  local dx, dy, dz = {0, 1, 1, 0, -1, -1}, {1, 0, -1, -1, 0, 1}, {-1, -1, 0, 1, 1, 0}      -- All six possible directions
  local cx, cy, cz = getcubecoord(x, y)
  local list = {}
  local i, a, b
  for i = 1, 6, 1 do
    a, b = getflatcoord(cx+dx[i], cy+dy[i], cz+dz[i])
    if tHEX[b][a].Player == 0 then table.insert(list, {X=a, Y=b}) end
  end
  return list
end

function getisadjacent(card1, card2)
  if getflatdist(card1.GX, card1.GY, card2.GX, card2.GY)==1 then return true
  else return false end
end

function getishomefree(pl)
  local free = false
  local i, j
  if pl == 1 then j = gGRID_HEIGHT
  else j = 1 end
  for i = 1, 5, 1 do
    if tHEX[j][i].Player == 0 then free = true end
  end
  return free
end

function checkvictory()
  local victory = false
  local i
  for i = 1, 5, 1 do
    if tHEX[1][i].Player == 1 then victory = true
    elseif tHEX[gGRID_HEIGHT][i].Player == 2 then victory = true
    else end
  end
  return victory
end

function AI()
  local backuptCLICK = deepcopy(tCLICK)   -- backup of what has been clicked
  local AIchoices = {}                    -- the choices I have for moves
  local AIchain_start = {}                -- the move that started this chain
  local AIlayer = 0                       -- counter for length of chain
  local report = ""
  local aiDECK, aiHAND, aiCOMBAT, aiGRAVE, aiDAYNIGHT, aiHEX
  local aiBUDGET, aiPOINTS, aiSACRIFICED, aiCEASEFIRE, aiACTION
  local a, b = gTURN, 3-gTURN             -- friendly & foe
  local i

  if AIloop < tSETTINGS.AIMaxLoops then   -- Am I still searching for moves?
    AIsimulation = true                   -- just testing, not for real!
    aiDECK = deepcopy(tDECK)              -- save the state of play
    aiHAND = deepcopy(tHAND)              -- save the state of play
    aiCOMBAT = deepcopy(tCOMBAT)          -- save the state of play
    aiGRAVE = deepcopy(tGRAVE)            -- save the state of play
    aiBUDGET = deepcopy(tBUDGET)          -- save the state of play
    aiPOINTS = deepcopy(tPOINTS)          -- save the state of play
    aiHEX = deepcopy(tHEX)                -- save the state of play
    aiSACRIFICED = gSACRIFICED            -- save the state of play
    aiCEASEFIRE = gCEASEFIRE              -- save the state of play
    aiACTION = gACTION                    -- save the state of play
    aiDAYNIGHT = gDAYNIGHT                -- save the state of play
    gTIMER = love.timer.getTime()
    repeat                                -- loop until this move chain comes to an end (or gets very long!)
      AIlayer = AIlayer + 1               -- counter for length of move chain
      AIchoices = listsomeAImoves()       -- list of moves I can make right now
      if next(AIchoices) ~= nil then  -- is there a move I can make?
        i = math.random(#AIchoices)  -- choose a move at random
        if AIlayer == 1 then AIchain_start = deepcopy(AIchoices[i]) end  -- remember the start
        report = AIreport(AIchoices[i])
        tryAImove(AIchoices[i][1], AIchoices[i][2], AIchoices[i][3], AIchoices[i][4], AIchoices[i][5], AIchoices[i][6], AIchoices[i][7], AIchoices[i][8])  -- do the move
        if AIchoices[i][9] > AIbest_score then  -- check if it's a new best score
          ailog("#"..AIloop..": NEW BEST TARGET MOVE...")
          ailog("-"..report)
          AIbest_score = AIchoices[i][9]  -- new best score
          AIbest_start = deepcopy(AIchain_start)  -- recall how I got here first move..!
        end
      end
    until next(AIchoices) == nil or AIlayer > 10 or love.timer.getTime()-gTIMER >(1/tSETTINGS.FPS)  -- until no choices / chain too long / takes too long!
    -- restore the state of play to what it was
    tDECK = deepcopy(aiDECK)
    tHAND = deepcopy(aiHAND)
    tCOMBAT = deepcopy(aiCOMBAT)
    tGRAVE = deepcopy(aiGRAVE)
    tBUDGET = deepcopy(aiBUDGET)
    tPOINTS = deepcopy(aiPOINTS)
    tHEX = deepcopy(aiHEX)
    gSACRIFICED = aiSACRIFICED
    gCEASEFIRE = aiCEASEFIRE
    gACTION = aiACTION
    gDAYNIGHT = aiDAYNIGHT
    AIloop = AIloop + 1  -- increase the counter towards making a move
  else  -- Time to move!
    AIsimulation = false  -- okay, we're not playing around anymore!
    if next(AIbest_start) ~= nil then  -- I've found something good to do, do the first bit!
      ailog(AIloop.." LOOPS COMPLETE. STARTING MOVE...")
      report = AIreport(AIbest_start)
      tryAImove(AIbest_start[1], AIbest_start[2], AIbest_start[3], AIbest_start[4], AIbest_start[5], AIbest_start[6], AIbest_start[7], AIbest_start[8])
      ailog("-"..report)
      ailog(tPOINTS[a].." points remaining.")
      AIbest_score = AIscore(a)  -- now current score is the one to beat
      AIbest_start = {}  -- and I have a blank slate to work with
      AIloop = 0  -- plenty of time to think again
      AIlastmove = love.timer.getTime()
    else  -- Nothing good found to do, end my turn
      gACTION = "End turn"
      ailog("No beneficial move left so end Player"..a.." turn")
      gTURN = b
      newturn(gTURN)
    end
  end
  if gTURN == a then tCLICK = deepcopy(backuptCLICK) end
end

function AIscore(pl)
  local a, b = pl, 3-pl         -- friendly & foe
  local i, j                    -- loop variables, coeffs
  local c, p, h, x               -- general variables

  score_budget = 0              -- Score for my budget vs their budget
  score_hand = 0                -- Score for cards in my hand
  score_allies = 0              -- Score for my allies vs their allies
  score_dist = 0                -- Score for placement of allies overall

  local AIbudget_ideal = 8      -- What's my ideal budget?
  local AIbudget_peak = 800     -- How much is that budget worth to me?
  local AIbudget_nominal = 5    -- How much is each budget point worth after that?
  local AIcard_minimum = 3      -- What's my lower card limit?
  local AIcard_peak = 900       -- How much is keeping at least that number of cards worth to me?
  local AIcard_nominal = 10     -- How much is each card worth after that?
  local AIcard_saver = 3        -- How far in advance will I wait for cards?
  local AIcard_penalty = 25     -- How much will I penalise overly expensive cards? (per point)
  local AIally_health = 0.9     -- How important is the health of allies?
  local AIally_power = 0.9      -- How important is the power of allies?
  local AIally_nominal = 100    -- How much is an ally's life worth to me?
  local AIally_defensive = 1    -- How much weight do I give to my defensive allies?
  local AIally_attacking = 1    -- How much weight do I give to their attacking allies?
  local AIdist_defend = 150     -- How much do I value keeping their allies at a distance?
  local AIdist_attack = 100     -- How much do I value moving my allies closer?
  local AIhome_bonus = 100      -- Bonus for each of my home spots that is occupied

  -- BUDGETS
  i=AIbudget_peak/(AIbudget_ideal^2)
  j=i*AIbudget_ideal*2
  c=tBUDGET[a]
  if c <= AIbudget_ideal then score_budget=score_budget+((j*c)-(i*(c^2)))
  else score_budget=score_budget+AIbudget_peak end
  score_budget=score_budget+(c*AIbudget_nominal)
  c=tBUDGET[b]
  if c <= AIbudget_ideal then score_budget=score_budget-((j*c)-(i*(c^2)))
  else score_budget=score_budget-AIbudget_peak end
  score_budget=score_budget-(c*AIbudget_nominal)
  score_budget=math.floor(score_budget)
  
  -- NUMBER OF CARDS IN HAND
  i=AIcard_peak/(AIcard_minimum^2)
  j=i*AIcard_minimum*2
  c=#tHAND[a]
  if c <= AIcard_minimum then score_hand=score_hand+((j*c)-(i*(c^2)))
  else score_hand=score_hand+AIcard_peak end
  score_hand=score_hand+(c*AIcard_nominal)
  c=#tHAND[b]
  if c <= AIcard_minimum then score_hand=score_hand-((j*c)-(i*(c^2)))
  else score_hand=score_hand-AIcard_peak end
  score_hand=score_hand-(c*AIcard_nominal)
  
  -- USEFULNESS OF CARD IN HAND & MULTIPLES
  for i = 1, #tHAND[a], 1 do
    j=getnumberinlist(tHAND[a][i].Name, tHAND[a])
    score_hand=score_hand-(25*(2^(j-1)))
    j=tHAND[a][i].Tier-tBUDGET[a]-AIcard_saver
    if j>0 then score_hand = score_hand-(j*AIcard_penalty) end
  end
  score_hand=math.floor(score_hand+500)  -- to counter negatives from duplicates
  
  -- DEFENSIVE ALLIES
  for i = 1, #tCOMBAT[a], 1 do
    score_allies = score_allies + AIally_nominal
    h = gethealth(tCOMBAT[a][i])
    p = getpower(tCOMBAT[a][i], a)
    if tCOMBAT[a][i].Effect[1]>0 then p=p*(0.75^(math.min(tCOMBAT[a][i].Effect[1],5))) end
    if tCOMBAT[a][i].Effect[2]>0 then p=p*1.1 end
    if tCOMBAT[a][i].Effect[3]>0 then p=p*(0.6^(math.min(tCOMBAT[a][i].Effect[3],5))) end
    if tCOMBAT[a][i].Effect[4]>0 then p=p*(0.9^(math.min(tCOMBAT[a][i].Effect[4],5))) end
    x = 0
    if tCOMBAT[a][i].Effect[11]>0 then x = x + 1 end
    if tCOMBAT[a][i].Effect[12]>0 then x = x + 3 end
    if tCOMBAT[a][i].Effect[13]>0 then x = x + 5 end
    h = h * (math.max(h-x, 0) / h)
    score_allies = score_allies + (((h^AIally_health)*(p^AIally_power)) * AIally_defensive)
  end
  
  -- ATTACKING ALLIES
  for i = 1, #tCOMBAT[b], 1 do
    score_allies = score_allies - AIally_nominal
    h = gethealth(tCOMBAT[b][i])
    p = getpower(tCOMBAT[b][i], b)
    if tCOMBAT[b][i].Effect[1]>0 then p=p*(0.75^(math.min(tCOMBAT[b][i].Effect[1],5))) end
    if tCOMBAT[b][i].Effect[2]>0 then p=p*1.1 end
    if tCOMBAT[b][i].Effect[3]>0 then p=p*(0.6^(math.min(tCOMBAT[b][i].Effect[3],5))) end
    score_allies = score_allies - (((h^AIally_health)*(p^AIally_power)) * AIally_attacking)
  end
  
  -- REDUCE SLIGHTLY(!) ALLIES EFFECT IF AT PEACE
  if gCEASEFIRE > 0 then score_allies = score_allies * 0.95 end
  score_allies=math.floor(score_allies)

  -- DEFENSIVE ALLIES DIST
  for i = 1, #tCOMBAT[a], 1 do
    if a==1 then 
      x = getflatdist(tCOMBAT[a][i].GX, tCOMBAT[a][i].GY, gGRID_MID, 1)
      if tCOMBAT[a][i].GX == gGRID_MID and tCOMBAT[a][i].GY == 1 then score_dist = score_dist + 10000 end
    else 
      x = getflatdist(tCOMBAT[a][i].GX, tCOMBAT[a][i].GY, gGRID_MID, gGRID_HEIGHT)
      if tCOMBAT[a][i].GX == gGRID_MID and tCOMBAT[a][i].GY == gGRID_HEIGHT then score_dist = score_dist + 10000 end
    end
    score_dist = score_dist + (AIdist_attack-(x^2))
  end
  
  for i = 1, 5, 1 do
    if a==1 and tHEX[gGRID_HEIGHT][i].Player==1 then score_dist = score_dist + AIhome_bonus
    elseif a==2 and tHEX[1][i].Player==2 then score_dist = score_dist + AIhome_bonus
    else end
  end
  
  -- ATTACKING ALLIES DIST
  for i = 1, #tCOMBAT[b], 1 do
    if b==1 then 
      x = getflatdist(tCOMBAT[b][i].GX, tCOMBAT[b][i].GY, gGRID_MID, 1)
      if tCOMBAT[b][i].GX == gGRID_MID and tCOMBAT[b][i].GY == 1 then score_dist = score_dist - 10000 end
    else 
      x = getflatdist(tCOMBAT[b][i].GX, tCOMBAT[b][i].GY, gGRID_MID, gGRID_HEIGHT) 
      if tCOMBAT[b][i].GX == gGRID_MID and tCOMBAT[b][i].GY == gGRID_HEIGHT then score_dist = score_dist - 10000 end
    end
    score_dist = score_dist - (AIdist_defend-(x^2))
  end

  return score_budget + score_hand + score_allies + score_dist
end

function listsomeAImoves()
  local list, move = {}, {}
  local a, b = gTURN, 3-gTURN
  local ha, co = 0, 0   -- hand_source, combat_source
  local ht = {0, 0}     -- hand_target
  local ct = {0, 0}     -- combat_target
  local click = {}
  local i, j, x, y

  -- Pick a random card from everywhere available (otherwise zero)
  if next(tHAND[a]) ~= nil then ha = math.random(#tHAND[a]) end
  if next(tCOMBAT[a]) ~= nil then co = math.random(#tCOMBAT[a]) end
  for i = 1, 2, 1 do
    if next(tHAND[i]) ~= nil then ht[i] = math.random(#tHAND[i]) end
    if next(tCOMBAT[i]) ~= nil then ct[i] = math.random(#tCOMBAT[i]) end
  end

  checkpoint()     -- save the state of play

  if ha ~= 0 and gSACRIFICED ~= true then   -- Sacrifice an ally
    gACTION = "Sacrifice"  
    move={"Sacrifice", ha, "Hand", 0, "None", 0, 0, 0}
    if tryAImove(move[1], move[2], move[3], move[4], move[5], move[6], move[7], move[8]) == true then
      table.insert(move, AIscore(a))
      table.insert(list, move)
    end  
    restorecheckpoint()
  end

  if ha ~= 0 then                           -- March an ally into battle
    if (tHAND[a][ha].Type == "Ally" or tHAND[a][ha].Type == "Hero") and tHAND[a][ha].Tier <= tPOINTS[a] and getishomefree(a) then
      gACTION = "Play to Combat"
      move={"Play to Combat", ha, "Hand", 0, "None", 0, 0, 0}
      if tryAImove(move[1], move[2], move[3], move[4], move[5], move[6], move[7], move[8]) == true then
        table.insert(move, AIscore(a))
        table.insert(list, move)
      end
      restorecheckpoint()
    end
  end

  if co ~= 0 then                           -- Move an ally to a free space
    x, y = math.random(gGRID_WIDTH), math.random(gGRID_HEIGHT)
    if getreach(tCOMBAT[a][co])>0 and tCOMBAT[a][co].Effect[1]==0 and tCOMBAT[a][co].Effect[3]==0 and tHEX[y][x].Player==0 then
      gACTION = "Move/Attack"
      move = {"Move/Attack", co, "Combat", 0, "None", 0, x, y}
      if tryAImove(move[1], move[2], move[3], move[4], move[5], move[6], move[7], move[8]) == true then
        table.insert(move, AIscore(a))
        table.insert(list, move)
      end
      restorecheckpoint()
    end
  end

  if co ~= 0 and gCEASEFIRE == 0 then       -- Attack a target using an ally
    if getreach(tCOMBAT[a][co])>0 and tCOMBAT[a][co].Effect[1]==0 and tCOMBAT[a][co].Effect[3]==0 and tCOMBAT[a][co].Effect[4]==0 and ct[b]~=0 then
      gACTION = "Move/Attack"
      move = {"Move/Attack", co, "Combat", ct[b], "Combat", b, tCOMBAT[b][ct[b]].GX, tCOMBAT[b][ct[b]].GY}
      if tryAImove(move[1], move[2], move[3], move[4], move[5], move[6], move[7], move[8]) == true then
        table.insert(move, AIscore(a))
        table.insert(list, move)
      end
      restorecheckpoint()
    end
  end

  if ha ~= 0 then   -- Play an action card from hand (on any target or none)
    if tHAND[a][ha].Type == "Action" and tHAND[a][ha].Tier <= tPOINTS[a] then
      gACTION = tHAND[a][ha].Spec1
      if getabilitytargets(gACTION) == 0 then
        move={"Play action", ha, "Hand", 0, "None", 0, 0, 0}
        if tryAImove(move[1], move[2], move[3], move[4], move[5], move[6], move[7], move[8]) == true then
          table.insert(move, AIscore(a))
          table.insert(list, move)
        end
      else
        click.Card = tHAND[a][ha]
        click.Source = "Hand"
        click.Player = a
        click.Index = ha
        table.insert(tCLICK, 1, deepcopy(click))  -- Need to click the card for findtargets to work!
        if findtargets() ~= false then
          tCLICK = {}                           -- Now unclick it because it gets clicked in tryAImove!
          harfway()
          for i = 1, 2, 1 do
            if ht[i] ~= 0 and tHAND[i][ht[i]].Target == true then
              move = {"Play action", ha, "Hand", ht[i], "Hand", i, 0, 0}
              if tryAImove(move[1], move[2], move[3], move[4], move[5], move[6], move[7], move[8]) == true then
                table.insert(move, AIscore(a))
                table.insert(list, move)
              end
              restoreharfway()
            end
            if ct[i] ~= 0 and tCOMBAT[i][ct[i]].Target == true then
              move = {"Play action", ha, "Hand", ct[i], "Combat", i, tCOMBAT[i][ct[i]].GX, tCOMBAT[i][ct[i]].GY}
              if tryAImove(move[1], move[2], move[3], move[4], move[5], move[6], move[7], move[8]) == true then
                table.insert(move, AIscore(a))
                table.insert(list, move)
              end
              restoreharfway()
            end
          end
        end
      end
    restorecheckpoint()
    end
  end

  if co ~= 0 then   -- Use a special ability belonging to an ally (on any target or none)
    for i = 1, 4, 1 do
      if tCOMBAT[a][co]["Spec"..i]~=nil and getrechargetime(tCOMBAT[a][co]["Spec"..i])>0 and tCOMBAT[a][co]["Charge"..i]==getrechargetime(tCOMBAT[a][co]["Spec"..i]) then
        gACTION = tCOMBAT[a][co]["Spec"..i]
        if getabilitytargets(gACTION) == 0 then
          move={"Spec"..i, co, "Combat", 0, "None", 0, 0, 0}
          if tryAImove(move[1], move[2], move[3], move[4], move[5], move[6], move[7], move[8]) == true then
            table.insert(move, AIscore(a))
            table.insert(list, move)
          end
        else
          click.Card = tCOMBAT[a][co]
          click.Source = "Combat"
          click.Player = a
          click.Index = co
          table.insert(tCLICK, 1, deepcopy(click))  -- Need to click the card for findtargets to work!
          if findtargets() ~= false then
            tCLICK = {}                             -- Now unclick it because it gets clicked in tryAImove!
            harfway()
            for j = 1, 2, 1 do
              if ct[j] ~= 0 then
                if tCOMBAT[j][ct[j]].Target == true then
                  move={"Spec"..i, co, "Combat", ct[j], "Combat", j, tCOMBAT[j][ct[j]].GX, tCOMBAT[j][ct[j]].GY}
                  if tryAImove(move[1], move[2], move[3], move[4], move[5], move[6], move[7], move[8]) == true then
                    table.insert(move, AIscore(a))
                    table.insert(list, move)
                  end
                  restoreharfway()
                end
              end
              if ht[j] ~= 0 then
                if tHAND[j][ht[j]].Target == true then
                  move={"Spec"..i, co, "Combat", ht[j], "Hand", j, tHAND[j][ht[j]].GX, tHAND[j][ht[j]].GY}
                  if tryAImove(move[1], move[2], move[3], move[4], move[5], move[6], move[7], move[8]) == true then
                    table.insert(move, AIscore(a))
                    table.insert(list, move)
                  end
                  restoreharfway()
                end
              end
            end
          end
        end
      end
      restorecheckpoint()
    end
  end

  return list
end

function tryAImove(move, origin, origin_home, target, target_home, target_player, target_x, target_y)
  local a, b = gTURN, 3-gTURN
  local r = false
  local origin_card, target_card = {}, {}
  local click = {}
  local i

  tCLICK = {}
  gACTION = move
  
  if origin_home == "Hand" and next(tHAND[a][origin]) ~= nil then origin_card = tHAND[a][origin]
  elseif origin_home == "Combat" and next(tCOMBAT[a][origin]) ~= nil then origin_card = tCOMBAT[a][origin]
  elseif origin_home == "Deck" and next(tDECK[a][origin]) ~= nil then origin_card = tDECK[a][origin]
  elseif origin_home == "Grave" and next(tGRAVE[a][origin]) ~= nil then origin_card = tGRAVE[a][origin]
  else end

  click.Card = origin_card
  click.Source = origin_home
  click.Player = a
  click.Index = origin
  table.insert(tCLICK, 1, deepcopy(click))

  if target > 0 then
    if target_home == "Hand" and next(tHAND[target_player][target]) ~= nil then target_card = tHAND[target_player][target]
    elseif target_home == "Combat" and next(tCOMBAT[target_player][target]) ~= nil then target_card = tCOMBAT[target_player][target]
    elseif target_home == "Deck" and next(tDECK[target_player][target]) ~= nil then target_card = tDECK[target_player][target]
    elseif target_home == "Grave" and next(tGRAVE[target_player][target]) ~= nil then target_card = tGRAVE[target_player][target]
    else end
    click = {}
    click.Card = target_card
    click.Source = target_home
    click.Player = target_player
    click.Index = target
    table.insert(tCLICK, 1, deepcopy(click))
  end

  if move == "Sacrifice" and next(origin_card) ~= nil then
    addpopup("P"..gTURN.." sacrifices "..origin_card.Name)
    sendcard("Grave", origin_card, origin_home, a, origin)
    tBUDGET[a] = tBUDGET[a] + 1
    tPOINTS[a] = tPOINTS[a] + 1
    r = true
    gSACRIFICED = true
  elseif move == "Play to Combat" and next(origin_card) ~= nil then
    addpopup("P"..gTURN.." plays "..origin_card.Name.." to combat")
    tPOINTS[a] = tPOINTS[a] - origin_card.Tier
    sendcard("Combat", origin_card, origin_home, a, origin)
    r = true
  elseif move == "Play action" and next(origin_card) ~= nil then
    gACTION = origin_card.Spec1
    if getabilitytargets(gACTION) == 0 then
      if playaction(origin_card, a) == true then
        tPOINTS[a] = tPOINTS[a] - origin_card.Tier
        sendcard("Grave", origin_card, origin_home, a, origin)
        r = true
      end
    elseif next(target_card) ~= nil then
      if playaction(origin_card, a) == true then
        tPOINTS[a] = tPOINTS[a] - origin_card.Tier
        dotarget()
        sendcard("Grave", origin_card, origin_home, a, origin)
        r = true
      end
    else end
  elseif string.sub(move, 1, 4)=="Spec" and next(origin_card) ~= nil then
    i = string.sub(move, 5, 5)
    gACTION = origin_card["Spec"..i]
    if getabilitytargets(gACTION) == 0 then
      if playaction(origin_card, a) == true then
        origin_card["Charge"..i] = 0
        r = true
      end
    elseif next(target_card) ~= nil then
      if playaction(origin_card, a) == true then
        origin_card["Charge"..i] = 0
        dotarget()
        r = true
      end
    else end
  elseif move == "Move/Attack" and next(origin_card) ~= nil and target == 0 then
    refreshmoves()
    if getreach(origin_card)>=tHEX[target_y][target_x].Moves then
      domove(target_x, target_y)
      r = true
    end
  elseif move == "Move/Attack" and next(origin_card) ~= nil and target ~= 0 then
    refreshmoves()
    if findtargets() then
      if target_card.Target==true then 
        dotarget()
        cleartargets()
        r = true
      end
    end
  else end

  tCLICK = {}

  return r
end

function AIreport(move)
  local report = ""
  local mycard, theircard = nil, nil
  local a, b = gTURN, move[6]  -- home player & target player (can be the same)

  if move[3] == "Hand" then mycard = tHAND[a][move[2]]
  elseif move[3] == "Combat" then mycard = tCOMBAT[a][move[2]]
  elseif move[3] == "Grave" then mycard = tGRAVE[a][move[2]]
  elseif move[3] == "Deck" then mycard = tDECK[a][move[2]]
  else mycard = nil end

  if move[4] > 0 then
    if move[5] == "Hand" then theircard = tHAND[b][move[4]]
    elseif move[5] == "Combat" then theircard = tCOMBAT[b][move[4]]
    elseif move[5] == "Grave" then theircard = tGRAVE[b][move[4]]
    elseif move[5] == "Deck" then theircard = tDECK[b][move[4]]
    else theircard = nil end
  end

  if move[1] == "Play to Combat" then report = "March "..mycard.Name..": "..move[9]
  elseif move[1] == "Play action" and theircard ~= nil then report = "Play action ("..mycard.Spec1..") using "..mycard.Name.." on "..theircard.Name..": "..move[9]
  elseif move[1] == "Play action" and theircard == nil then report = "Play action ("..mycard.Spec1..") using "..mycard.Name..": "..move[9]
  elseif move[1] == "Move/Attack" and theircard == nil then report = "Move "..mycard.Name.." to ("..move[7]..","..move[8].."): "..move[9]
  elseif move[1] == "Move/Attack" and theircard ~= nil then report = "Attack "..theircard.Name.." using "..mycard.Name..": "..move[9]
  elseif move[1] == "Sacrifice" then report = "Sacrifice "..mycard.Name..": "..move[9]
  elseif string.sub(move[1], 1, 4)=="Spec" and theircard ~= nil then report = "Use ability ("..mycard[move[1]]..") using "..mycard.Name.. "on "..theircard.Name..": "..move[9]
  elseif string.sub(move[1], 1, 4)=="Spec" and theircard == nil then report = "Use ability ("..mycard[move[1]]..") using "..mycard.Name..": "..move[9]
  else report = "Not sure what this move was!?" end

  return report
end

function checkpoint()
  bDECK = deepcopy(tDECK)
  bHAND = deepcopy(tHAND)
  bCOMBAT = deepcopy(tCOMBAT)
  bGRAVE = deepcopy(tGRAVE)
  bBUDGET = deepcopy(tBUDGET)
  bPOINTS = deepcopy(tPOINTS)
  bHEX = deepcopy(tHEX)
  bSACRIFICED = gSACRIFICED
  bCEASEFIRE = gCEASEFIRE
  bACTION = gACTION
  bDAYNIGHT = gDAYNIGHT
end

function restorecheckpoint()
  tDECK = deepcopy(bDECK)
  tHAND = deepcopy(bHAND)
  tCOMBAT = deepcopy(bCOMBAT)
  tGRAVE = deepcopy(bGRAVE)
  tBUDGET = deepcopy(bBUDGET)
  tPOINTS = deepcopy(bPOINTS)
  tHEX = deepcopy(bHEX)
  gSACRIFICED = bSACRIFICED
  gCEASEFIRE = bCEASEFIRE
  gACTION = bACTION
  gDAYNIGHT = bDAYNIGHT
end

function harfway()
  hDECK = deepcopy(tDECK)
  hHAND = deepcopy(tHAND)
  hCOMBAT = deepcopy(tCOMBAT)
  hGRAVE = deepcopy(tGRAVE)
  hBUDGET = deepcopy(tBUDGET)
  hPOINTS = deepcopy(tPOINTS)
  hHEX = deepcopy(tHEX)
  hSACRIFICED = gSACRIFICED
  hCEASEFIRE = gCEASEFIRE
  hACTION = gACTION
  hDAYNIGHT = gDAYNIGHT
end

function restoreharfway()
  tDECK = deepcopy(hDECK)
  tHAND = deepcopy(hHAND)
  tCOMBAT = deepcopy(hCOMBAT)
  tGRAVE = deepcopy(hGRAVE)
  tBUDGET = deepcopy(hBUDGET)
  tPOINTS = deepcopy(hPOINTS)
  tHEX = deepcopy(hHEX)
  gSACRIFICED = hSACRIFICED
  gCEASEFIRE = hCEASEFIRE
  gACTION = hACTION
  gDAYNIGHT = hDAYNIGHT
end

function addpopup(msg)
  if AIsimulation==false then
    if msg ~= "Choose target" and msg ~= "Choose targets" and msg ~= "Choose location" then writelog(msg) end
    gPOPUP_COUNT = tSETTINGS.FPS*tSETTINGS.PopUpTime
    gPOPUP_MSG = msg
  end
end

function addfloatytext(card, text)
  entry = {}
  entry.Text = text
  entry.X, entry.Y = card.X, card.Y
  entry.Timer = 0
  table.insert(card.FloatyText, entry)
end

function writelog(text)
  if not love.filesystem.getInfo("gamelogs", "directory") then love.filesystem.createDirectory("gamelogs") end
  if gGAMELOG == nil then gGAMELOG = love.filesystem.newFile("gamelogs/"..os.date("%Y-%m-%d %H%M.txt")) end
  gGAMELOG:open("a")
  gGAMELOG:write(text.."\r\n")
  gGAMELOG:close()
end

function ailog(text)
  local log
  if not love.filesystem.getInfo("gamelogs/ai.txt", "file") then
    log = love.filesystem.newFile("gamelogs/ai.txt")
    log:open("w")
    log:write(os.date("%Y-%m-%d %H:%M\r\n"))
    log:close()
  end
  log = love.filesystem.newFile("gamelogs/ai.txt")
  log:open("a")
  log:write(text.."\r\n")
  log:close()
end

function gridtoscreen(x, y)
  local cols, rows = gGRID_WIDTH, gGRID_HEIGHT
  local i, j = xGRID, yGRID
  i = i + ((x-((cols+1)/2))*tSETTINGS.HexWidth)
  if y%2 == 0 then i = i + (tSETTINGS.HexWidth/2) end
  j = j + ((y-((rows+1)/2))*(tSETTINGS.HexHeight*0.75))
  return i, j 
end

function screentogrid(x, y)
  local gx, gy = 0, 0
  local cols, rows = gGRID_WIDTH, gGRID_HEIGHT
  local i, j, cx, cy
  for j = 1, rows, 1 do
    for i = 1, cols, 1 do
      cx, cy = gridtoscreen(i, j)
      if getwithinhex(x, y, cx, cy, tSETTINGS.HexWidth, tSETTINGS.HexHeight) then gx, gy = i, j end
    end
  end
  return gx, gy
end

function endgame()
  local winner, loser = 0, 0
  local i, j
  gFADE = 255
  gGAMEMODE = "menu"
  if tHEX[1][gGRID_MID].Player==1 then winner, loser = 1, 2
  elseif tHEX[gGRID_HEIGHT][gGRID_MID].Player==2 then winner, loser = 2, 1
  else end
  if winner > 0 and loser > 0 then
    if tCONTROL[winner]=="Player" and tCONTROL[loser]=="AI" then
      gENDING = tSETTINGS.FPS * 5
      bonuscard(winner, loser)
      if gLEVEL == tPLAYER.LEVEL and gDIFFICULTY>=2 and tPLAYER.LEVEL < 11 then 
        gNEWLEVEL = tPLAYER.LEVEL + 1
        tPLAYER.LEVEL = gNEWLEVEL
      end
      savedeck(tCOLLECTION, "collection.txt")
      gGAMEMODE = "endscreen"
    else
      iBACKGROUND = love.graphics.newImage("backgrounds/background12.jpg")
      playtrack(2)
    end
    if tCONTROL[2]=="AI" then
      if winner == 1 then tPLAYER.WIN = tPLAYER.WIN + 1
      else tPLAYER.LOSE = tPLAYER.LOSE + 1 end
      writesingletable("player.txt", tPLAYER)
    end
  end
end

function bonuscard(winner, loser)
  local list = {}
  local i, j, alreadylisted
  tDECK[loser] = loaddeck("ai/deck"..tostring((3*(gLEVEL-1))+gDIFFICULTY)..".txt")
  if next(tDECK[loser])~=nil then
    for i = 1, #tDECK[loser], 1 do
      alreadylisted = false
      if next(list)~=nil then
        for j = 1, #list, 1 do
          if tDECK[loser][i].Name == list[j] then alreadylisted = true end
        end
      end
      if alreadylisted == false then
        if (tDECK[loser][i].Type == "Ally" or tDECK[loser][i].Type == "Action") and getnumberinlist(tDECK[loser][i].Name, tCOLLECTION)<4 then table.insert(list, tDECK[loser][i].Name)
        elseif tDECK[loser][i].Type == "Hero" and getnumberinlist(tDECK[loser][i].Name, tCOLLECTION)==0 then table.insert(list, tDECK[loser][i].Name)
        else end
      end
    end
  end
  if next(list)~=nil and next(tBLUEPRINT)~=nil then
    i = math.random(#list)
    for j = 1, #tBLUEPRINT, 1 do
      if list[i]==tBLUEPRINT[j].Name then 
        table.insert(tCOLLECTION, deepcopy(tBLUEPRINT[j]))
        gBONUSCARD = j
      end
    end
  end
end